"""
Views for MyTardis storage box records
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_storage_box(storage_box, render_format):
    """
    Render storage box

    :param storage_box: The storage box to be rendered.
    :type storage_box: :class:`mtclient.models.storagebox.StorageBox`
    """
    if render_format == 'json':
        return render_storage_box_as_json(storage_box)
    return render_storage_box_as_table(storage_box)


def render_storage_box_as_json(storage_box, indent=2, sort_keys=True):
    """
    Returns JSON representation of storage_box.

    :param storage_box: The storage box to be rendered.
    :type storage_box: :class:`mtclient.models.storagebox.StorageBox`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        storage_box.response_dict, indent=indent, sort_keys=sort_keys)


def render_storage_box_as_table(storage_box):
    """
    Returns ASCII table view of storage_box.
    """
    storage_box_options_attributes = ""

    table = Texttable(max_width=0)
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["StorageBox field", "Value"])
    table.add_row(["ID", storage_box.id])
    table.add_row(["Name", storage_box.name])
    table.add_row(["Description", storage_box.description])
    table.add_row(["Django Storage Class", storage_box.django_storage_class])
    table.add_row(["Max Size", storage_box.max_size])
    table.add_row(["Status", storage_box.status])
    storage_box_options_attributes += table.draw() + "\n"

    if storage_box.options:
        storage_box_options_attributes += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l'])
        table.set_cols_valign(['m', 'm'])
        table.header(["StorageBoxOption Key", "StorageBoxOption Value"])
        for option in storage_box.options:
            table.add_row([option.key, option.value])
        storage_box_options_attributes += table.draw() + "\n"

    if storage_box.attributes:
        storage_box_options_attributes += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l'])
        table.set_cols_valign(['m', 'm'])
        table.header(["StorageBoxAttribute Key", "StorageBoxAttribute Value"])
        for attribute in storage_box.attributes:
            table.add_row([attribute.key, attribute.value])
        storage_box_options_attributes += table.draw() + "\n"

    return storage_box_options_attributes


def render_storage_boxes(storage_boxes, render_format, display_heading=True):
    """
    Render storage boxes.

    :param storage_boxes: The `ResultSet` of storage boxes to be rendered.
    :type storage_boxes: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_storage_boxes_as_json(storage_boxes)
    return render_storage_boxes_as_table(storage_boxes, display_heading)


def render_storage_boxes_as_json(storage_boxes, indent=2, sort_keys=True):
    """
    Returns JSON representation of storage_boxes.

    :param storage_boxes: The result set of storage boxes to be displayed.
    :type storage_boxes: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        storage_boxes.response_dict, indent=indent, sort_keys=sort_keys)


def render_storage_boxes_as_table(storage_boxes, display_heading=True):
    """
    Returns ASCII table view of storage_boxes.

    :param storage_boxes: The storage boxes to be rendered.
    :type storage_boxes: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: StorageBox\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (storage_boxes.url, storage_boxes.total_count,
           storage_boxes.limit,
           storage_boxes.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Name", "Description"])
    for storage_box in storage_boxes:
        table.add_row([storage_box.id, storage_box.name,
                       storage_box.description])
    return heading + table.draw() + "\n"



#!/usr/bin/env python
"""
client.py
"""
from __future__ import print_function

import os
import sys
import logging
import logging.config

from . import __version__ as VERSION
from .models.config import Config
from .models.config import DEFAULT_CONFIG_PATH
from .controllers.api import ApiController
from .controllers.config import ConfigController
from .controllers.facility import FacilityController
from .controllers.instrument import InstrumentController
from .controllers.experiment import ExperimentController
from .controllers.dataset import DatasetController
from .controllers.datafile import DataFileController
from .controllers.storagebox import StorageBoxController
from .controllers.schema import SchemaController
from .argparser import ArgParser


def run():
    """
    Main function for command-line interface.
    """
    # pylint: disable=too-many-branches

    args = ArgParser().get_args()

    if args.model == 'version':
        print("MyTardis Client v%s" % VERSION)
        sys.exit(0)
    if args.verbose and (not hasattr(args, 'json') or not args.json):
        print("MyTardis Client v%s" % VERSION)

    config_path = DEFAULT_CONFIG_PATH
    if not os.path.exists(config_path) or \
            args.model == 'config':
        ConfigController(config_path).configure(args)
        if args.model == 'config':
            sys.exit(0)
    config = Config(config_path)
    config.validate()

    logging.config.fileConfig(config.logging_config_path,
                              disable_existing_loggers=False)
    logging.getLogger("requests").setLevel(logging.WARNING)
    if args.version:
        print("MyTardis Client v%s" % VERSION)

    if args.verbose and (not hasattr(args, 'json') or not args.json):
        print("General config: %s" % config_path)
        print("Logging config: %s" % config.logging_config_path)
        print("MyTardis URL: %s" % config.url)
        print("Username: %s" % config.username)

    if args.model == 'api':
        ApiController().run_command(args)
    elif args.model == 'facility':
        FacilityController().run_command(args)
    elif args.model == 'instrument':
        InstrumentController().run_command(args)
    elif args.model == 'experiment':
        ExperimentController().run_command(args)
    elif args.model == 'dataset':
        DatasetController().run_command(args)
    elif args.model == 'datafile':
        DataFileController().run_command(args)
    elif args.model == 'storagebox':
        StorageBoxController().run_command(args)
    elif args.model == 'schema':
        SchemaController().run_command(args)


if __name__ == "__main__":
    run()



"""
Views for MyTardis api endpoints.
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_api_schema(api_schema, render_format):
    """
    Render API schema

    :param api_schema: The API schema model to be displayed.
    :type api_schema: :class:`mtclient.models.api.ApiSchema`
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if render_format == 'json':
        return render_api_schema_as_json(api_schema)
    return render_api_schema_as_table(api_schema)


def render_api_schema_as_json(api_schema, indent=2, sort_keys=True):
    """
    Returns JSON representation of API schema.

    :param api_schema: The API schema model to be displayed.
    :type api_schema: :class:`mtclient.models.api.ApiSchema`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        api_schema.response_dict, indent=indent, sort_keys=sort_keys)


def render_api_schema_as_table(api_schema):
    """
    Returns ASCII table view of API schema.

    :param api_schema: The API schema model to be displayed.
    :type api_schema: :class:`mtclient.models.api.ApiSchema`
    """
    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(["t", "t"])
    table.header(["API Schema field", "Value"])
    table.add_row(["Model", api_schema.model])
    table.add_row(["Fields",
                   "\n".join(sorted([field for field in api_schema.fields]))])
    table.add_row(["Filtering",
                   json.dumps(api_schema.filtering, indent=2, sort_keys=True)])
    table.add_row(["Ordering",
                   json.dumps(api_schema.ordering, indent=2, sort_keys=True)])
    return table.draw() + "\n"


def render_api_endpoints(api_endpoints, render_format, display_heading=True):
    """
    Render API endpoints

    :param api_endpoints: The API endpoints to be rendered.
    :type api_endpoints: :class:`mtclient.models.api.ApiEndpoints`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for
        an `ApiEndpoints` set, setting `display_heading` to True
        ensures that a heading is displayed before the results table.
        The heading includes the URL resolved to perform the query.
    """
    if render_format == 'json':
        return render_api_endpoints_as_json(api_endpoints)
    return render_api_endpoints_as_table(api_endpoints, display_heading)


def render_api_endpoints_as_json(api_endpoints, indent=2, sort_keys=True):
    """
    Returns JSON representation of api_endpoints.

    :param api_endpoints: The API endpoints to be rendered.
    :type api_endpoints: :class:`mtclient.models.api.ApiEndpoints`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        api_endpoints.response_dict, indent=indent, sort_keys=sort_keys)


def render_api_endpoints_as_table(api_endpoints, display_heading=True):
    """
    Returns ASCII table view of api_endpoints.

    :param api_endpoints: The API endpoints to be rendered.
    :type api_endpoints: :class:`mtclient.models.api.ApiEndpoints`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for
        an `ApiEndpoints` set, setting `display_heading` to True
        ensures that a heading is displayed before the results table.
        The heading includes the URL resolved to perform the query.
    """
    heading = "\n" \
        "API Endpoints\n" if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["l", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["Model", "List Endpoint", "Schema"])
    for api_endpoint in api_endpoints:
        table.add_row([api_endpoint.model, api_endpoint.list_endpoint,
                       api_endpoint.schema])
    return heading + table.draw() + "\n"



"""
test_storagebox.py

Tests for functionality to query the StorageBox model via
the MyTardis REST API's facility resource
"""
import json
import requests_mock

from mtclient.conf import config
from mtclient.models.storagebox import StorageBox


def test_storagebox_list():
    """
    Test listing storage box records exposed in a MyTardis REST API
    """
    mock_storagebox_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "attributes": [
                ],
                "description": "Storage box description",
                "django_storage_class":
                    "tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage",
                "id": 1,
                "max_size": "9999999999",
                "name": "local box at /home/mytardis/var/local",
                "options": [
                    {
                        "id": 1,
                        "key": "location",
                        "resource_uri": "/api/v1/storageboxoption/1/",
                        "storage_box": "/api/v1/storagebox/1/",
                        "value": "/home/mytardis/var/local",
                        "value_type": "string"
                    }
                ],
                "resource_uri": "/api/v1/storagebox/1/",
                "status": "online"
            }
        ]
    }
    mock_list_response = json.dumps(mock_storagebox_list)
    with requests_mock.Mocker() as mocker:
        list_storageboxes_url = "%s/api/v1/storagebox/?format=json" % config.url
        mocker.get(list_storageboxes_url, text=mock_list_response)
        boxes = StorageBox.objects.all()
        boxes._execute_query()
        assert boxes._result_set.response_dict == mock_storagebox_list


def test_storagebox_get():
    """
    Test getting a storage box record by ID
    """
    mock_storagebox = {
        "attributes": [
        ],
        "description": "Storage box description",
        "django_storage_class": "tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage",
        "id": 1,
        "max_size": "9999999999",
        "name": "local box at /home/mytardis/var/local",
        "options": [
            {
                "id": 1,
                "key": "location",
                "resource_uri": "/api/v1/storageboxoption/1/",
                "storage_box": "/api/v1/storagebox/1/",
                "value": "/home/mytardis/var/local",
                "value_type": "string"
            }
        ],
        "resource_uri": "/api/v1/storagebox/1/",
        "status": "online"
    }
    mock_get_response = json.dumps(mock_storagebox)
    with requests_mock.Mocker() as mocker:
        get_storagebox_url = "%s/api/v1/storagebox/1/?format=json" % config.url
        mocker.get(get_storagebox_url, text=mock_get_response)
        storagebox = StorageBox.objects.get(id=1)
        assert storagebox.response_dict == mock_storagebox



"""
test_instrument.py

Tests for functionality to query the Instrument model via
the MyTardis REST API's instrument resource
"""
import json
import requests_mock

from mtclient.conf import config
from mtclient.models.instrument import Instrument


def test_instrument_list():
    """
    Test listing instrument records exposed in a MyTardis REST API
    """
    mock_instrument_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "created_time": None,
                "facility": {
                    "created_time": None,
                    "id": 1,
                    "manager_group": {
                        "id": 1,
                        "name": "test-facility-managers",
                        "resource_uri": "/api/v1/group/1/"
                    },
                    "modified_time": None,
                    "name": "Test Facility",
                    "resource_uri": "/api/v1/facility/1/"
                },
                "id": 1,
                "modified_time": None,
                "name": "Test Instrument",
                "resource_uri": "/api/v1/instrument/1/"
            }
        ]
    }
    mock_list_response = json.dumps(mock_instrument_list)
    with requests_mock.Mocker() as mocker:
        list_instruments_url = "%s/api/v1/instrument/?format=json&facility__id=1" % config.url
        mocker.get(list_instruments_url, text=mock_list_response)
        instruments = Instrument.list(filters="facility__id=1")
        assert instruments.response_dict == mock_instrument_list


def test_instrument_get():
    """
    Test getting a instrument record by ID
    """
    mock_instrument = {
        "created_time": None,
        "facility": {
            "created_time": None,
            "id": 1,
            "manager_group": {
                "id": 1,
                "name": "test-facility-managers",
                "resource_uri": "/api/v1/group/1/"
            },
            "modified_time": None,
            "name": "Test Facility",
            "resource_uri": "/api/v1/facility/1/"
        },
        "id": 1,
        "modified_time": None,
        "name": "Test Instrument",
        "resource_uri": "/api/v1/instrument/1/"
    }
    mock_get_response = json.dumps(mock_instrument)
    with requests_mock.Mocker() as mocker:
        get_instrument_url = "%s/api/v1/instrument/1/?format=json" % config.url
        mocker.get(get_instrument_url, text=mock_get_response)
        instrument = Instrument.objects.get(id=1)
        assert instrument.response_dict == mock_instrument



"""
Controller class for listing API endpoints.
"""
from __future__ import print_function

from mtclient.models.api import ApiEndpoint
from mtclient.models.api import ApiSchema
from mtclient.utils import get_render_format
from mtclient.views import render


class ApiController(object):
    """
    Controller class for listing API endpoints.
    """
    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        render_format = get_render_format(args)
        if command == "list":
            return self.list(args, render_format)
        if command == "get":
            return self.get(args, render_format)
        raise NotImplementedError("Invalid command: %s" % args.command)

    def list(self, _args, render_format):
        """
        Display list of API endpoints.
        """
        # pylint: disable=no-self-use
        api_endpoints = ApiEndpoint.list()
        print(render(api_endpoints, render_format))

    def get(self, args, render_format):
        """
        Display schema for a model's API endpoint.
        """
        # pylint: disable=no-self-use
        api_schema = ApiSchema.get(args.api_model)
        print(render(api_schema, render_format))



"""
test_facility_cli.py

Tests for querying the MyTardis REST API's facility endpoints
via the command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.facility import FacilityController


def test_facility_list_cli_json(capfd):
    """
    Test listing facilities, requesting output in JSON format
    """
    mock_facility_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "manager_group": {
                    "id": 1,
                    "name": "test-facility-managers",
                    "resource_uri": "/api/v1/group/1/"
                },
                "name": "Test Facility",
                "resource_uri": "/api/v1/facility/1/"
            }
        ]
    }
    mock_facility_list_response = json.dumps(mock_facility_list)
    with requests_mock.Mocker() as mocker:
        facility_list_url = "%s/api/v1/facility/?format=json" % config.url
        mocker.get(facility_list_url, text=mock_facility_list_response)
        facility_controller = FacilityController()
        args = Namespace(
            model='facility', command='list', json=True, verbose=False,
            limit=None, offset=None, order_by=None)

        facility_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_facility_list

        facility_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_facility_list

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'facility', 'list', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_facility_list
        sys.argv = sys_argv


def test_facility_list_cli_table(capfd):
    """
    Test listing facility records, requesting output in ASCII table format
    """
    mock_facility_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "manager_group": {
                    "id": 1,
                    "name": "test-facility-managers",
                    "resource_uri": "/api/v1/group/1/"
                },
                "name": "Test Facility",
                "resource_uri": "/api/v1/facility/1/"
            }
        ]
    }
    mock_facility_list_response = json.dumps(mock_facility_list)
    expected = textwrap.dedent("""
        Model: Facility
        Query: %s/api/v1/facility/?format=json
        Total Count: 1
        Limit: 20
        Offset: 0

        +----+---------------+------------------------+
        | ID |     Name      |     Manager Group      |
        +====+===============+========================+
        |  1 | Test Facility | test-facility-managers |
        +----+---------------+------------------------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        facility_list_url = "%s/api/v1/facility/?format=json" % config.url
        mocker.get(facility_list_url, text=mock_facility_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'facility', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_facility_get_cli_json(capfd):
    """
    Test looking up and displaying a facility via the command-line interface
    """
    mock_facility = {
        "id": 1,
        "manager_group": {
            "id": 1,
            "name": "test-facility-managers",
            "resource_uri": "/api/v1/group/1/"
        },
        "name": "Test Facility",
        "resource_uri": "/api/v1/facility/1/"
    }
    mock_facility_get_response = json.dumps(mock_facility)
    with requests_mock.Mocker() as mocker:
        get_facility_url = "%s/api/v1/facility/1/?format=json" % config.url
        mocker.get(get_facility_url, text=mock_facility_get_response)
        facility_controller = FacilityController()
        args = Namespace(
            model='facility', command='get', facility_id=1, json=True,
            verbose=False)
        facility_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_facility


def test_facility_get_cli_table(capfd):
    """
    Test getting facility record via the command-line interface,
    requesting output in ASCII table format
    """
    mock_facility = {
        "id": 1,
        "manager_group": {
            "id": 1,
            "name": "test-facility-managers",
            "resource_uri": "/api/v1/group/1/"
        },
        "name": "Test Facility",
        "resource_uri": "/api/v1/facility/1/"
    }
    mock_facility_get_response = json.dumps(mock_facility)
    mock_instrument_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "created_time": None,
                "facility": {
                    "created_time": None,
                    "id": 1,
                    "manager_group": {
                        "id": 1,
                        "name": "test-facility-managers",
                        "resource_uri": "/api/v1/group/1/"
                    },
                    "modified_time": None,
                    "name": "Test Facility",
                    "resource_uri": "/api/v1/facility/1/"
                },
                "id": 1,
                "modified_time": None,
                "name": "Test Instrument",
                "resource_uri": "/api/v1/instrument/1/"
            }
        ]
    }
    mock_instruments_response = json.dumps(mock_instrument_list)
    expected = textwrap.dedent("""
        Model: Facility

        +----------------+------------------------+
        | Facility field |         Value          |
        +================+========================+
        | ID             | 1                      |
        +----------------+------------------------+
        | Name           | Test Facility          |
        +----------------+------------------------+
        | Manager Group  | test-facility-managers |
        +----------------+------------------------+


        Model: Instrument
        Query: %s/api/v1/instrument/?format=json&facility__id=1
        Total Count: 1
        Limit: 20
        Offset: 0

        +----+-----------------+---------------------------+
        | ID |      Name       |         Facility          |
        +====+=================+===========================+
        |  1 | Test Instrument | <Facility: Test Facility> |
        +----+-----------------+---------------------------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        get_facility_url = "%s/api/v1/facility/1/?format=json" % config.url
        mocker.get(get_facility_url, text=mock_facility_get_response)
        instruments_url = "%s/api/v1/instrument/?format=json&facility__id=1" % config.url
        mocker.get(instruments_url, text=mock_instruments_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'facility', 'get', '1']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv



"""
Controller class for running commands (list, get)
on schemas.
"""
from __future__ import print_function

from ..models.schema import Schema
from ..views import render

from .cli import ModelCliController


class SchemaController(ModelCliController):
    """
    Controller class for running commands (list, get) on schema records.
    """
    def __init__(self):
        super(SchemaController, self).__init__()
        self.allowed_commands = ["list", "get"]
        self.primary_key_arg = "schema_id"
        self.model = Schema

    def get(self, args, render_format):
        """
        Display a single record
        """
        primary_key = getattr(args, "schema_id")
        instance = Schema.objects.get(id=primary_key, param_names=args.params)
        print(render(instance, render_format))



"""
argparser/dataset.py
"""
import textwrap


def build_dataset_parser(argument_parser):
    """
    Builds parsing rules for dataset-related command-line interface arguments.
    """
    # pylint: disable=too-many-locals
    dataset_help = \
        "Display a list of dataset records or a single dataset record."
    dataset_usage = "mytardis dataset [-h] {list,get,create,update,download} ..."
    dataset_parser = \
        argument_parser.model_parsers.add_parser("dataset",
                                                 help=dataset_help,
                                                 usage=dataset_usage)
    dataset_command_parsers = \
        dataset_parser.add_subparsers(help='available commands',
                                      dest='command')

    dataset_list_help = "Display a list of dataset records."
    dataset_list_usage = textwrap.dedent("""\
        mytardis dataset list
            [--exp EXP] [--limit LIMIT] [--offset OFFSET] [--order_by ORDER_BY] [--json]
            [--filter FILTER]

          EXAMPLE

          $ mytardis dataset list --exp 20

          Model: Dataset
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/dataset/?format=json&experiments__id=20
          Total Count: 4
          Limit: 20
          Offset: 0

          +------------+------------------------+------------------------+------------+
          | Dataset ID |     Experiment(s)      |      Description       | Instrument |
          +============+========================+========================+============+
          |         34 | /api/v1/experiment/20/ | James Test Dataset 001 | None       |
          +------------+------------------------+------------------------+------------+
          |         33 | /api/v1/experiment/20/ | James Test Dataset 003 | None       |
          +------------+------------------------+------------------------+------------+
          |         32 | /api/v1/experiment/20/ | James Test Dataset 002 | None       |
          +------------+------------------------+------------------------+------------+
          |         31 | /api/v1/experiment/20/ | James Test Dataset 001 | None       |
          +------------+------------------------+------------------------+------------+
        """)
    dataset_command_list_parser = \
        dataset_command_parsers.add_parser("list", help=dataset_list_help,
                                           usage=dataset_list_usage)
    dataset_command_list_parser.add_argument("--exp",
                                             help="The experiment ID.")
    dataset_command_list_parser.add_argument(
        "--limit", help="Maximum number of results to return.")
    dataset_command_list_parser.add_argument(
        "--offset",
        help="Skip this many records from the start of the result set.")
    dataset_command_list_parser.add_argument(
        "--order_by",
        help="Order by this field.")
    dataset_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")
    dataset_command_list_parser.add_argument(
        "--filter", help="Filter on these fields, e.g. \"description=Dataset Description\".")

    dataset_get_help = "Display a single dataset record."
    dataset_get_usage = textwrap.dedent("""\
        mytardis dataset get [-h] [--json] dataset_id

          EXAMPLE

          Suppose we want to retrieve a dataset record with dataset_id=35.
          The "mytardis dataset get 35" command will also display any
          dataset parameters associated with the dataset and any
          datafiles associated with it.  The datafile query could result
          in pagination.  If "Total Count" (below) is greater
          than "Limit", then there are additional pages of datafiles
          not shown in the dataset view below.

          $ mytardis dataset get 35
          +---------------+------------------------+
          | Dataset field |         Value          |
          +===============+========================+
          | ID            | 35                     |
          +---------------+------------------------+
          | Experiment(s) | /api/v1/experiment/24/ |
          +---------------+------------------------+
          | Description   | James Test Dataset 004 |
          +---------------+------------------------+
          | Instrument    | None                   |
          +---------------+------------------------+

          +---------------------+---------------------+------------------------+-------------------------+-----------------+----------------+---------+
          | DatasetParameter ID |       Schema        |     Parameter Name     |      String Value       | Numerical Value | Datetime Value | Link ID |
          +=====================+=====================+========================+=========================+=================+================+=========+
          |                   1 | Test dataset schema | Example Parameter Name | Example Parameter Value |                 |                |         |
          +---------------------+---------------------+------------------------+-------------------------+-----------------+----------------+---------+


          Model: DataFile
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/dataset_file/?format=json&dataset__id=35
          Total Count: 2
          Limit: 20
          Offset: 0

          +--------------+-----------+-----------+-------------------------------------+----------+-----------+----------------------------------+
          | DataFile ID  | Directory | Filename  |                 URI                 | Verified |   Size    |             MD5 Sum              |
          +==============+===========+===========+=====================================+==========+===========+==================================+
          |          120 |           | hello.txt | James Test Dataset 004-35/hello.txt | True     |  25 bytes | 214e52fcf7f98d9fb8588b42cfb7987f |
          +--------------+-----------+-----------+-------------------------------------+----------+-----------+----------------------------------+
          |          121 |           | test.txt  | James Test Dataset 004-35/test.txt  | False    |   8 bytes | 74119cb1c9568b96feb7fa392006e40f |
          +--------------+-----------+-----------+-------------------------------------+----------+-----------+----------------------------------+
        """)
    dataset_command_get_parser = \
        dataset_command_parsers.add_parser("get", help=dataset_get_help,
                                           usage=dataset_get_usage)
    dataset_command_get_parser.add_argument("dataset_id",
                                            help="The dataset ID.")
    dataset_command_get_parser.add_argument(
        "--metadata", action='store_true', help="Include metadata.")
    dataset_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    dataset_create_help = "Create a dataset record."
    dataset_create_usage = textwrap.dedent("""\
        mytardis dataset create
            [--instrument INSTRUMENT] [--params PARAMS] experiment_id description

          EXAMPLE

          First let's look up a dataset schema we can use to add some
          metadata to the new experiment record.

          $ mytardis schema list

          +----+---------------------------+-----------------------------------------------------------+-------------------+---------+-----------+--------+
          | ID |           Name            |                         Namespace                         |       Type        | Subtype | Immutable | Hidden |
          +====+===========================+===========================================================+===================+=========+===========+========+
           ...  ...                         ...                                                         ...                 ...       ...         ...
          +----+---------------------------+-----------------------------------------------------------+-------------------+---------+-----------+--------+
          | 13 | Sample Dataset Schema     | https://mytardis.org/schemas/sample-dataset-schema        | Dataset schema    |         | False     | False  |
          +----+---------------------------+-----------------------------------------------------------+-------------------+---------+-----------+--------+

          "mytardis schema get <schema_id>" displays the parameters
          associated with this schema:

          $ mytardis schema get 13
          +--------------+----------------------------------------------------+
          | Schema field |                       Value                        |
          +==============+====================================================+
          | ID           | 13                                                 |
          +--------------+----------------------------------------------------+
          | Name         | Sample Dataset Schema                              |
          +--------------+----------------------------------------------------+
          | Namespace    | https://mytardis.org/schemas/sample-dataset-schema |
          +--------------+----------------------------------------------------+
          | Type         | Dataset schema                                     |
          +--------------+----------------------------------------------------+
          | Subtype      |                                                    |
          +--------------+----------------------------------------------------+
          | Immutable    | False                                              |
          +--------------+----------------------------------------------------+
          | Hidden       | False                                              |
          +--------------+----------------------------------------------------+

          +------------------+-----------------------+-----------------------+-----------+-------+-----------+---------------+-------+---------+-----------------+
          | ParameterName ID |       Full Name       |         Name          | Data Type | Units | Immutable | Is Searchable | Order | Choices | Comparison Type |
          +==================+=======================+=======================+===========+=======+===========+===============+=======+=========+=================+
          |               34 | Sample Parameter Name | sample_parameter_name | String    |       | False     | False         | 9999  |         | Exact value     |
          +------------------+-----------------------+-----------------------+-----------+-------+-----------+---------------+-------+---------+-----------------+

          Now that we know the schema's namespace, the parameter name(s) and
          their data type(s), we can create some dataset metadata.

          Below we use cat to create the metadata file, but you can use
          any text editor.

          $ cat << EOF > params.json
          [
              {
                  "schema": "https://mytardis.org/schemas/sample-dataset-schema",
                  "parameters": [
                      {
                          "name": "sample_parameter_name",
                          "value": "Sample Parameter Value"
                      }
                  ]
              }
          ]
          EOF

          Now we create the dataset record (and associate it with experiment
          ID 20) , using the metadata file we just created:

          $ mytardis dataset create --params params.json 20 "Dataset With Params"
          +---------------+------------------------+
          | Dataset field |         Value          |
          +===============+========================+
          | ID            | 39                     |
          +---------------+------------------------+
          | Experiment(s) | /api/v1/experiment/20/ |
          +---------------+------------------------+
          | Description   | Dataset With Params    |
          +---------------+------------------------+
          | Instrument    | None                   |
          +---------------+------------------------+

          +---------------------+-----------------------+-----------------------+------------------------+-----------------+----------------+---------+
          | DatasetParameter ID |        Schema         |    Parameter Name     |      String Value      | Numerical Value | Datetime Value | Link ID |
          +=====================+=======================+=======================+========================+=================+================+=========+
          |                   3 | Sample Dataset Schema | Sample Parameter Name | Sample Parameter Value |                 |                |         |
          +---------------------+-----------------------+-----------------------+------------------------+-----------------+----------------+---------+

          Dataset created successfully.
                """)
    dataset_command_create_parser = \
        dataset_command_parsers.add_parser("create",
                                           help=dataset_create_help,
                                           usage=dataset_create_usage)
    dataset_command_create_parser.add_argument(
        "experiment_id", help="The experiment ID.")
    dataset_command_create_parser.add_argument(
        "description", help="The dataset description.")
    dataset_command_create_parser.add_argument("--instrument",
                                               help="The instrument ID.")
    dataset_command_create_parser.add_argument(
        "--params", help="A JSON file containing dataset parameters.")

    dataset_update_help = "Update a dataset record."
    dataset_update_usage = textwrap.dedent("""\
        mytardis dataset update
            [--description DESCRIPTION] dataset_id

          EXAMPLE
          $ mytardis dataset update --description "Renamed Dataset" 39
          HTTP 401
          Traceback (most recent call last):
            ...
          Exception

          Dataset updates are not yet enabled in the MyTardis API.
          See the "update_detail" method in tardis/tardis_portal/api.py
        """)
    dataset_cmd_update_parser = \
        dataset_command_parsers.add_parser("update",
                                           help=dataset_update_help,
                                           usage=dataset_update_usage)
    dataset_cmd_update_parser.add_argument(
        "dataset_id", help="The ID of the dataset to update.")
    dataset_cmd_update_parser.add_argument(
        "--description", help="The new description of the dataset.")

    dataset_download_help = "Download a dataset."
    dataset_download_usage = textwrap.dedent("""\
        mytardis dataset download dataset_id

          EXAMPLE

          $ mytardis dataset download 99
          Downloaded: hello.txt
            """)
    dataset_cmd_download_parser = \
        dataset_command_parsers.add_parser("download",
                                           help=dataset_download_help,
                                           usage=dataset_download_usage)
    dataset_cmd_download_parser.add_argument("dataset_id",
                                             help="The dataset ID.")



"""
test_experiment_cli.py

Tests for querying the MyTardis REST API's experiment endpoints
via the command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.experiment import ExperimentController


def test_experiment_list_cli_json(capfd):
    """
    Test listing experiments, requesting output in JSON format
    """
    mock_experiment_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "title": "Test Experiment",
                "description": "exp description",
                "authors": [
                ],
                "created_by": "/api/v1/user/1/",
                "created_time": "2017-08-03T12:59:48.600082",
                "institution_name": "Monash University",
                "owner_ids": [
                    1,
                ],
                "parameter_sets": [
                ],
                "public_access": 1,
                "update_time": "2017-08-03T12:59:48.600106",
                "resource_uri": "/api/v1/experiment/1/",
            }
        ]
    }
    mock_experiment_list_response = json.dumps(mock_experiment_list)
    with requests_mock.Mocker() as mocker:
        experiment_list_url = "%s/api/v1/experiment/?format=json" % config.url
        mocker.get(experiment_list_url, text=mock_experiment_list_response)
        exp_controller = ExperimentController()
        args = Namespace(
            model='experiment', command='list', json=True, verbose=False,
            filter=None, limit=None, offset=None, order_by=None)

        exp_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_experiment_list

        exp_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_experiment_list

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'experiment', 'list', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_experiment_list
        sys.argv = sys_argv


def test_experiment_list_cli_table(capfd):
    """
    Test listing experiment records, requesting output in ASCII table format
    """
    mock_experiment_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "title": "Test Experiment",
                "description": "exp description",
                "authors": [
                ],
                "created_by": "/api/v1/user/1/",
                "created_time": "2017-08-03T12:59:48.600082",
                "institution_name": "Monash University",
                "owner_ids": [
                    1,
                ],
                "parameter_sets": [
                ],
                "public_access": 1,
                "update_time": "2017-08-03T12:59:48.600106",
                "resource_uri": "/api/v1/experiment/1/",
            }
        ]
    }
    mock_experiment_list_response = json.dumps(mock_experiment_list)
    expected = textwrap.dedent("""
        Model: Experiment
        Query: %s/api/v1/experiment/?format=json&order_by=-created_time
        Total Count: 1
        Limit: 20
        Offset: 0

        +----+-------------------+-----------------+
        | ID |    Institution    |      Title      |
        +====+===================+=================+
        |  1 | Monash University | Test Experiment |
        +----+-------------------+-----------------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        experiment_list_url = "%s/api/v1/experiment/?format=json" % config.url
        mocker.get(experiment_list_url, text=mock_experiment_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'experiment', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_experiment_get_cli_json(capfd):
    """
    Test looking up and displaying an experiment via the command-line interface
    """
    mock_experiment = {
        "id": 1,
        "title": "Test Experiment",
        "description": "exp description",
        "authors": [
        ],
        "created_by": "/api/v1/user/1/",
        "created_time": "2017-08-03T12:59:48.600082",
        "institution_name": "Monash University",
        "owner_ids": [
            1,
        ],
        "parameter_sets": [
        ],
        "public_access": 1,
        "update_time": "2017-08-03T12:59:48.600106",
        "resource_uri": "/api/v1/experiment/1/",
    }
    mock_experiment_get_response = json.dumps(mock_experiment)
    with requests_mock.Mocker() as mocker:
        get_experiment_url = "%s/api/v1/experiment/1/?format=json" % config.url
        mocker.get(get_experiment_url, text=mock_experiment_get_response)
        exp_controller = ExperimentController()
        args = Namespace(
            model='experiment', command='get', experiment_id=1, json=True,
            metadata=False, verbose=False)
        exp_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_experiment


def test_experiment_get_cli_table(capfd):
    """
    Test getting experiment record via the command-line interface,
    requesting output in ASCII table format
    """
    mock_experiment = {
        "id": 1,
        "title": "Test Experiment",
        "description": "exp description",
        "authors": [
        ],
        "created_by": "/api/v1/user/1/",
        "created_time": "2017-08-03T12:59:48.600082",
        "institution_name": "Monash University",
        "owner_ids": [
            1,
        ],
        "parameter_sets": [
            {
                "id": 1,
                "experiment": "/api/v1/experiment/1/",
                "schema": {
                    "id": 1,
                    "name": "Series Metadata",
                    "namespace": "http://schema/namespace",
                    "hidden": False,
                    "immutable": True,
                    "subtype": "",
                    "type": 1
                },
                "parameters": [
                    {
                        "id": 1,
                        "name": "/api/v1/parametername/1/",
                        "parameterset": "/api/v1/experimentparameter/1/",
                        "resource_uri": "/api/v1/experimentparameter/1/",
                        "string_value": "param value",
                        "numerical_value": None,
                        "datetime_value": None,
                        "link_id": None,
                        "value": None
                    }
                ]
            }
        ],
        "public_access": 1,
        "update_time": "2017-08-03T12:59:48.600106",
        "resource_uri": "/api/v1/experiment/1/",
    }
    mock_schema = {
        "hidden": False,
        "id": 1,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": "",
        "type": 1
    }
    mock_schema_response = json.dumps(mock_schema)
    mock_pname = {
        "id": 1,
        "choices": "",
        "comparison_type": 1,
        "data_type": 2,
        "full_name": "Parameter Name",
        "immutable": True,
        "is_searchable": False,
        "name": "param name",
        "order": 1,
        "resource_uri": "/api/v1/parametername/1/",
        "schema": "/api/v1/schema/1/",
        "units": ""
    }
    mock_pname_response = json.dumps(mock_pname)
    mock_experiment_get_response = json.dumps(mock_experiment)
    mock_dataset_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 0
        },
        "objects": [
        ]
    }
    mock_ds_list_response = json.dumps(mock_dataset_list)
    expected = textwrap.dedent("""
        +------------------+-------------------+
        | Experiment field |       Value       |
        +==================+===================+
        | ID               | 1                 |
        +------------------+-------------------+
        | Institution      | Monash University |
        +------------------+-------------------+
        | Title            | Test Experiment   |
        +------------------+-------------------+
        | Description      | exp description   |
        +------------------+-------------------+
              
        +------------------------+-------------+----------------+--------------+-----------------+----------------+---------+
        | ExperimentParameter ID |   Schema    | Parameter Name | String Value | Numerical Value | Datetime Value | Link ID |
        +========================+=============+================+==============+=================+================+=========+
        |                      1 | Schema Name | param name     | param value  |                 |                |         |
        +------------------------+-------------+----------------+--------------+-----------------+----------------+---------+
    """)
    with requests_mock.Mocker() as mocker:
        get_experiment_url = "%s/api/v1/experiment/1/?format=json" % config.url
        mocker.get(get_experiment_url, text=mock_experiment_get_response)
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_schema_response)
        get_pname_url = "%s/api/v1/parametername/1/?format=json" % config.url
        mocker.get(get_pname_url, text=mock_pname_response)
        ds_list_url = "%s/api/v1/dataset/?format=json&experiments__id=1" % config.url
        mocker.get(ds_list_url, text=mock_ds_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'experiment', 'get', '1', '--metadata']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv



"""
Model class for MyTardis API v1's DataFileResource.
"""
from __future__ import print_function

import mimetypes
import json
import os
import cgi
import hashlib
import logging
from datetime import datetime

import requests
from six.moves import urllib

from ..conf import config
from ..utils import extend_url, add_filters
from ..utils.exceptions import DuplicateKey
from .model import Model
from .resultset import ResultSet
from .schema import Schema
from .schema import ParameterName

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def md5_sum(file_path, blocksize=65536):
    """
    Calculate MD5 checksum without reading the whole file into memory.
    """
    hasher = hashlib.md5()
    with open(file_path, 'rb') as datafile:
        buf = datafile.read(blocksize)
        while buf:
            hasher.update(buf)
            buf = datafile.read(blocksize)
        return hasher.hexdigest()


class DataFile(Model):
    """
    Model class for MyTardis API v1's DataFileResource.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict, include_metadata=False):
        from .replica import Replica

        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.dataset = response_dict['dataset']
        self.directory = response_dict['directory'] or ""
        self.filename = response_dict['filename']
        self.size = response_dict['size']
        self.md5sum = response_dict['md5sum']
        self.replicas = []
        for replica_json in response_dict['replicas']:
            self.replicas.append(Replica(replica_json))
        self.parameter_sets = []
        if include_metadata:
            for datafile_param_set_json in response_dict['parameter_sets']:
                self.parameter_sets.append(
                    DataFileParameterSet(datafile_param_set_json))

    def __str__(self):
        """
        Return a string representation of a datafile
        """
        return "<%s: %s>" % (
            type(self).__name__, os.path.join(self.directory, self.filename))

    @property
    def verified(self):
        """
        All replicas (DFOs) must be verified and there must be
        at least one replica (DFO).
        """
        if not self.replicas:
            return False
        for replica in self.replicas:
            if not replica.verified:
                return False
        return True

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of datafiles.

        :param filters: Filters, e.g. "dataset__id=123&filename=file1.txt"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`DataFile` records.
        """
        url = "%s/api/v1/dataset_file/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(DataFile, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single datafile record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the DataFile to retrieve

        :return: A :class:`DataFile` record.

        :raises requests.exceptions.HTTPError:
        """
        datafile_id = kwargs.get("id")
        if not datafile_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for DataFile get "
                "at this stage.")
        include_metadata = kwargs.get("include_metadata", False)
        url = "%s/api/v1/dataset_file/%s/?format=json" % \
            (config.url, datafile_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return DataFile(response.json(), include_metadata=include_metadata)

    @staticmethod
    def create(dataset_id, storagebox, dataset_path, path,
               check_local_paths=True, create_dataset_symlink=True):
        """
        Create one or more DataFile records, depending on whether the
        supplied path is a single file or a directory.

        :param dataset_id: The ID of the dataset to create the datafile
            record(s) in.
        :param storagebox: The storage box containing the datafile(s).
        :param dataset_path:
            The path to the directory which is to be mapped to a MyTardis
            dataset, meaning that files immediately inside this directory
            will have DataFile records created with the "directory" field
            set to "", and files within dataset_path/subdir1/ will have
            DataFile records creatd with the "directory" field set to "subdir1"
            etc.
            The MyTardis Client will create a symlink to dataset_path in
            ~/.config/mytardisclient/servers/[mytardis_hostname]/ which
            will enable MyTardis to verify and ingest the file if the
            filesystem running mytardisclient is accessible to MyTardis,
            e.g. by SSHFS.
        :param path: The path to a file to be represented in the DataFile
            record or to a directory containing the files to create records
            for.  The subdirectory ('subdir1') to be
            recorded in the DataFile record(s) will be determined
            automatically by compareing the dataset_path with the path.
        """
        if not dataset_path:
            raise Exception("The dataset_path argument is required.")
        if check_local_paths and not os.path.exists(path):
            raise Exception("The path doesn't exist: %s" % path)
        if os.path.isdir(path):
            return DataFile.create_datafiles(dataset_id, storagebox,
                                             dataset_path, path)
        return DataFile.create_datafile(
            dataset_id, storagebox, dataset_path, path,
            check_local_paths=check_local_paths,
            create_dataset_symlink=create_dataset_symlink)

    @staticmethod
    def create_datafiles(dataset_id, storagebox, dataset_path, dir_path):
        """
        Create a DataFile record for each file within the dir_path directory.

        :param dataset_id: The ID of the dataset to create the datafile
            record(s) in.
        :param storagebox: The storage box containing the datafile(s).
        :param dataset_path:
            The path to the directory which is to be mapped to a MyTardis
            dataset, meaning that files immediately inside this directory
            will have DataFile records created with the "directory" field
            set to "", and files within dataset_path/subdir1/ will have
            DataFile records creatd with the "directory" field set to "subdir1"
            etc.
            The MyTardis Client will create a symlink to dataset_path in
            ~/.config/mytardisclient/servers/[mytardis_hostname]/ which
            will enable MyTardis to verify and ingest the file if the
            filesystem running mytardisclient is accessible to MyTardis,
            e.g. by SSHFS.
        :param dir_path: The path to a directory containing file(s) to
            create DataFile records for.  If the dir_path is relative (not
            absolute) path, e.g. 'dataset1/subdir1', then the MyTardis Client
            will assume that the first component of the path (e.g. 'dataset1/')
            is the local dataset path, and create a symlink to this path in
            ~/.config/mytardisclient/servers/[mytardis_hostname]/
            which will enable MyTardis to verify and ingest its file(s).
            If dir_path is an absolute path, e.g.
            '/home/james/dataset1/subdir1/', then the dataset_path
            argument must be used to specified the dataset path, e.g.
            '/home/james/dataset1'.
        """
        num_datafiles_created = 0

        def log_error(err):
            """
            Log an error if os.listdir(...) fails during os.walk(...)
            """
            logger.error(str(err))

        for root, _, files in os.walk(dir_path, onerror=log_error):
            for filename in files:
                file_path = os.path.join(root, filename)
                try:
                    DataFile.create_datafile(dataset_id, storagebox,
                                             dataset_path, file_path,
                                             return_new_datafile=False)
                    num_datafiles_created += 1
                except DuplicateKey:
                    logger.warning("A DataFile record already exists for %s",
                                   file_path)
        return num_datafiles_created

    @staticmethod
    def create_datafile(dataset_id, storagebox, dataset_path, file_path,
                        return_new_datafile=True, check_local_paths=True,
                        create_dataset_symlink=True,
                        size=None, md5sum=None, mimetype=None):
        """
        Create a DataFile record.

        :param dataset_id: The ID of the dataset to create the datafile in.
        :param storagebox: The storage box containing the datafile.
        :param dataset_path:
            The path to the directory which is to be mapped to a MyTardis
            dataset, meaning that files immediately inside this directory
            will have DataFile records created with the "directory" field
            set to "", and files within dataset_path/subdir1/ will have
            DataFile records creatd with the "directory" field set to "subdir1"
            etc.
            The MyTardis Client will create a symlink to dataset_path in
            ~/.config/mytardisclient/servers/[mytardis_hostname]/ which
            will enable MyTardis to verify and ingest the file if the
            filesystem running mytardisclient is accessible to MyTardis,
            e.g. by SSHFS.
        :param file_path: The local path to the file to be represented in
            the DataFile record.  The subdirectory ('subdir1') to be
            recorded in the DataFile record(s) will be determined
            automatically by compareing the dataset_path with the file_path.

        :return: A new :class:`DataFile` record.

        See also: :func:`mtclient.models.datafile.DataFile.upload`

        Suppose someone with username james generates a file called
        "results.dat" on a data analysis server called analyzer.example.com
        in the directory ~james/analysis/dataset1/.  User james could grant
        the MyTardis server temporary access to his account on
        analyzer.example.com and then have MyTardis copy the file(s) into
        a more permanent location.

        If james agrees to allow the MyTardis server to do so, it could
        SSHFS-mount james@analyzer.example.com:/home/james/analysis/,
        e.g. at /mnt/sshfs/james-anaylzer/

        Then user james doesn't need to upload results.dat, he just needs to
        tell MyTardis how to access it, and tell MyTardis that it is not yet
        in a permanent location.

        MyTardis's default storage box model generates datafile object
        identifiers which include a dataset description (e.g. 'dataset1')
        and a unique ID, resulting in path like 'dataset1-123/results.dat'
        for the datafile object.  Because user james doesn't want to have
        to create the 'dataset1-123' folder himself, he could entrust the
        MyTardis Client to do it for him.

        The MyTardis administrator can create a storage box for james called
        "james-analyzer" which is of type "receiving", meaning that it is a
        temporary location.  The storage box record (which only needs to be
        accessed by the MyTardis administrator) would include a StorageBoxOption
        with key 'location' and value '/mnt/sshfs/james-analyzer'.

        Once james knows the dataset ID of the dataset he wants to upload to
        (123 in this case), he can create a DataFile record as follows:

        mytardis datafile create 123 --storagebox=james-analyzer ~/analysis/dataset1/results.dat

        The file_path argument (set to ~/analysis/dataset1/results.dat)
        specifies the location of 'results.dat' on the analysis server.

        To enable the MyTardis server to access (and verify) the file via
        SSHFS / SFTP, a symbolic link can be created in
        ~james/.mytardisclient/servers/[mytardis_hostname]/, named "dataset1-123" pointing to
        the location of 'results.dat', i.e. ~james/analysis/dataset1/.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        from .dataset import Dataset

        if not dataset_path:
            raise Exception("The dataset_path argument is required.")
        if check_local_paths and not os.path.exists(file_path):
            raise Exception("Path doesn't exist: %s" % file_path)
        if check_local_paths and os.path.isdir(file_path):
            raise Exception("The path should be a single file: %s" % file_path)
        dataset = Dataset.objects.get(id=dataset_id)
        local_dataset_path = dataset_path
        file_path_without_dataset = os.path.relpath(file_path,
                                                    dataset_path)
        (directory, filename) = os.path.split(file_path_without_dataset)

        uri = os.path.join("%s-%s" % (dataset.description, dataset_id),
                           directory, filename)
        if create_dataset_symlink:
            dataset_symlink_path = \
                os.path.join(config.datasets_path,
                             "%s-%s" % (dataset.description, dataset_id))
            if not os.path.exists(dataset_symlink_path):
                print("Creating symlink to: %s in "
                      "~/.config/mytardisclient/servers/%s/ called %s"
                      % (local_dataset_path, config.hostname,
                         "%s-%s" % (dataset.description, dataset_id)))
                os.symlink(os.path.abspath(local_dataset_path),
                           os.path.join(config.datasets_path,
                                        "%s-%s" % (dataset.description,
                                                   dataset_id)))
        if DataFile.exists(dataset_id, directory, filename):
            if directory and directory != "":
                _file_path = os.path.join(directory, filename)
            else:
                _file_path = filename
            raise DuplicateKey("A DataFile record already exists for file "
                               "'%s' in dataset ID %s." % (_file_path,
                                                           dataset_id))
        if not size:
            size = str(os.stat(file_path).st_size)
        if not md5sum:
            md5sum = md5_sum(file_path)
        if not mimetype:
            mimetype = mimetypes.guess_type(file_path)[0]
        replicas = [{
            "url": uri,
            "location": storagebox,
            "protocol": "file",
            "verified": False
        }]
        new_datafile_json = {
            'dataset': "/api/v1/dataset/%s/" % dataset_id,
            'filename': filename,
            'directory': directory or "",
            'md5sum': md5sum,
            'size': size,
            'mimetype': mimetype,
            'replicas': replicas,
            'parameter_sets': []
        }
        url = "%s/api/v1/dataset_file/" % config.url
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_datafile_json))
        response.raise_for_status()
        logger.info("Created a DataFile record for %s", file_path)
        if return_new_datafile:
            datafile_id = response.headers['location'].split("/")[-2]
            new_datafile = DataFile.objects.get(id=datafile_id)
            return new_datafile
        return None

    @staticmethod
    def download(datafile_id, basedir=None, overwrite=False,
                 force_overwrite=False):
        """
        Download datafile with id datafile_id

        :param datafile_id: The ID of a datafile to download.
        :param basedir: If specified, the datafile will be downloaded to
                        the path obtained by joining basedir with the
                        DataFile's directory field.
        :param overwrite: If set to True, existing files will be re-downloaded
                          and overwritten without asking for confirmation if
                          their file size is wrong.
        :param force_overwrite: If set to True, existing files will be
                                re-downloaded and overwritten without asking
                                for confirmation, even if their file size is
                                correct.
        """
        from clint.textui import progress  # pylint: disable=import-error

        url = "%s/api/v1/dataset_file/%s/download/" \
            % (config.url, datafile_id)
        headers = {
            "Authorization": "ApiKey %s:%s" % (config.username,
                                               config.apikey)}
        response = requests.get(url=url, headers=headers, stream=True)
        response.raise_for_status()
        datafile = DataFile.objects.get(id=datafile_id)
        try:
            _, params = cgi.parse_header(
                response.headers.get('Content-Disposition', ''))
            try:
                filename = params['filename']
            except KeyError:
                # If the download request is redirected to S3,
                # we may not have a filename header in the response,
                # so we can look up the datafile's filename instead:
                filename = datafile.filename
        except KeyError:
            print("response.headers: %s" % response.headers)
            raise
        filepath = filename
        if basedir:
            path = os.path.join(basedir, datafile.directory)
            if not os.path.exists(path):
                os.makedirs(path)
            filepath = os.path.join(path, datafile.filename)
        if os.path.exists(filepath) and not force_overwrite:
            if os.path.getsize(filepath) == datafile.size:
                logger.warning(
                    "Not re-downloading %s because its size is correct.",
                    filepath)
                return
            if not overwrite:
                from ..utils.confirmation import query_yes_no
                if not query_yes_no("Overwrite '%s'?" % filepath):
                    return
        with open(filepath, 'wb') as fileobj:
            total_length = int(response.headers.get('content-length'))
            # Hide progress bar for small files:
            if total_length < 10000000:  # 10 MB
                hide = True
            else:
                hide = None  # Leave it up to client.textui.progress
            chunk_size = 1000000
            for chunk in progress.bar(
                    response.iter_content(chunk_size=chunk_size),
                    hide=hide,
                    label="Downloading: %s " % filepath,
                    expected_size=(total_length / chunk_size) + 1):
                # filter out keep-alive new chunks:
                if chunk:
                    fileobj.write(chunk)
                    fileobj.flush()
            if hide:
                print("Downloaded: %s" % filepath)

    @staticmethod
    def upload(dataset_id, storagebox, dataset_path, file_path):
        """
        Upload datafile to dataset with ID dataset_id,
        using HTTP POST.

        :param dataset_id: The ID of the dataset to create the datafile in.
        :param dataset_path:
            The path to the directory which is to be mapped to a MyTardis
            dataset, meaning that files immediately inside this directory
            will have DataFile records created with the "directory" field
            set to "", and files within dataset_path/subdir1/ will have
            DataFile records creatd with the "directory" field set to "subdir1"
            etc.
            The MyTardis Client will create a symlink to dataset_path in
            ~/.config/mytardisclient/servers/[mytardis_hostname]/ which
            will enable MyTardis to verify and ingest the file if the
            filesystem running mytardisclient is accessible to MyTardis,
            e.g. by SSHFS.
        :param file_path: The local path to the file to be represented in
            the DataFile record.  If dataset_path is not specified,
            file_path must be a relative (not absolute) path, e.g.
            'dataset1/subdir1/datafile1.txt'.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        if not dataset_path:
            raise Exception("The dataset_path argument is required.")
        if not os.path.exists(file_path):
            raise Exception("Path doesn't exist: %s" % file_path)
        url = "%s/api/v1/dataset_file/" % config.url
        created_time = datetime.fromtimestamp(
            os.stat(file_path).st_ctime).isoformat()
        file_path_without_dataset = os.path.relpath(file_path,
                                                    dataset_path)
        directory, filename = os.path.split(file_path_without_dataset)
        if DataFile.exists(dataset_id, directory, filename):
            if directory and directory != "":
                _file_path = os.path.join(directory, filename)
            else:
                _file_path = filename
            raise DuplicateKey("A DataFile record already exists for file "
                               "'%s' in dataset ID %s." % (_file_path,
                                                           dataset_id))
        md5sum = md5_sum(file_path)
        file_data = {"dataset": "/api/v1/dataset/%s/" % dataset_id,
                     "filename": filename,
                     "directory": directory,
                     "md5sum": md5sum,
                     "size": str(os.stat(file_path).st_size),
                     "mimetype": mimetypes.guess_type(file_path)[0],
                     "created_time": created_time}
        if storagebox:
            file_data['replicas'] = [
                {
                    "url": "",
                    "protocol": "file",
                    "location": storagebox
                }
            ]
        file_obj = open(file_path, 'rb')
        headers = {
            "Authorization": "ApiKey %s:%s" % (config.username,
                                               config.apikey)}
        response = requests.post(url, headers=headers,
                                 data={"json_data": json.dumps(file_data)},
                                 files={'attached_file': file_obj})
        file_obj.close()
        response.raise_for_status()
        if directory:
            print("Uploaded: %s/%s" % (directory, file_path))
        else:
            print("Uploaded: %s" % file_path)

    @staticmethod
    def update(datafile_id, md5sum):
        """
        Update a DataFile record.

        :param datafile_id: The ID of a datafile to be updated.
        :param md5sum: The new MD5 sum value.

        This method is not usable yet, because the MyTardis API doesn't yet
        allow update_detail to be performed on DataFile records.

        For a large file, its upload can commence before the local MD5 sum
        calculation is complete, i.e.  the DataFile record can be initially
        created with a bogus checksum which is later updated using this
        method.
        """
        updated_fields_json = {'md5sum': md5sum}
        url = "%s/api/v1/dataset_file/%s/" % \
            (config.url, datafile_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        response.raise_for_status()
        datafile_json = response.json()
        return DataFile(datafile_json)

    @staticmethod
    def verify(datafile_id):
        """
        Ask MyTardis to verify a datafile with id datafile_id

        :param datafile_id: The ID of a datafile to be verified.
        """
        url = "%s/api/v1/dataset_file/%s/verify/" \
            % (config.url, datafile_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        print("Requested verification of datafile ID %s." % datafile_id)

    @staticmethod
    def exists(dataset_id, directory, filename):
        """
        If MyTardis is running with DEBUG=False, then we won't
        be able detect duplicate key errors easily, we will just
        receive a generic HTTP 500 from the MyTardis API. This
        method checks whether a DataFile record already exists
        for the supplied dataset_id, directory and filename.

        :param dataset_id: The ID of the dataset to check existence in.
        :param directory: The directory within the dataset to check existence in.
        :param filename: The filename to check for existence.

        :return: True if a matching DataFile record already exists.
        """

        url = "%s/api/v1/dataset_file/?format=json" % config.url
        url += "&dataset__id=%s" % dataset_id
        url += "&filename=%s" % urllib.parse.quote(filename)
        if directory and directory != "":
            url += "&directory=%s" % urllib.parse.quote(directory)
        response = requests.get(url=url, headers=config.default_headers)
        logger.debug("GET %s %s", url, response.status_code)
        if response.status_code < 200 or response.status_code >= 300:
            raise Exception("Failed to check for existing file '%s' "
                            "in dataset ID %s." % (filename, dataset_id))
        return response.json()['meta']['total_count'] > 0


class DataFileParameterSet(object):
    """
    Model class for MyTardis API v1's DataFileParameterSetResource.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.datafile = response_dict['datafile']
        self.schema = Schema(response_dict['schema'])
        self.parameters = []
        for datafile_param_json in response_dict['parameters']:
            self.parameters.append(DataFileParameter(datafile_param_json))

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of datafile parameter sets

        :param filters: Filters, e.g. "datafiles__id=12345"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`DatasetParameterSet` records,
            encapsulated in a `ResultSet` object`.
        """
        url = "%s/api/v1/datafileparameterset/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(DataFileParameterSet, url, response.json())


class DataFileParameter(object):
    """
    Model class for MyTardis API v1's DataFileParameterResource.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict):
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        pname_id = response_dict['name'].split('/')[-2]
        self.name = ParameterName.objects.get(id=pname_id)
        self.string_value = response_dict['string_value']
        self.numerical_value = response_dict['numerical_value']
        self.datetime_value = response_dict['datetime_value']
        self.link_id = response_dict['link_id']
        self.value = response_dict['value']

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        List datafile parameter records in parameter set.

        :param datafile_param_set: The datafile parameter set to
            list parameters for.
        """



"""
Controller class for running commands (list, get, create, update)
on dataset records.
"""
from __future__ import print_function

from mtclient.models.dataset import Dataset
from mtclient.views import render

from .cli import ModelCliController


class DatasetController(ModelCliController):
    """
    Controller class for running commands (list, get, create, update)
    on dataset records.
    """
    def __init__(self):
        super(DatasetController, self).__init__()
        self.allowed_commands = ["list", "get", "create", "update", "download"]
        self.primary_key_arg = "dataset_id"
        self.model = Dataset

    def list(self, args, render_format):
        """
        Display list of dataset records.
        """
        # pylint: disable=no-self-use
        if args.exp:
            filters = "experiments__id=%s" % args.exp
        else:
            filters = ""
        if args.filter:
            filters += "&%s" % args.filter
        datasets = Dataset.list(
            filters, args.limit, args.offset, args.order_by)
        print(render(datasets, render_format))

    def get(self, args, render_format):
        """
        Display dataset record.
        """
        # pylint: disable=no-self-use
        dataset = Dataset.objects.get(
            id=args.dataset_id, include_metadata=args.metadata)
        print(render(dataset, render_format))
        if render_format == 'table':
            from mtclient.models.datafile import DataFile
            datafiles = DataFile.list("dataset__id=%s" % args.dataset_id)
            if datafiles:
                print(render(datafiles, render_format))

    def create(self, args, render_format):
        """
        Create dataset record.
        """
        # pylint: disable=no-self-use
        dataset = Dataset.create(
            args.experiment_id, args.description, args.instrument, args.params)
        print(render(dataset, render_format))
        print("Dataset created successfully.")

    def update(self, args, render_format):
        """
        Update dataset record.
        """
        # pylint: disable=no-self-use
        dataset = Dataset.update(args.dataset_id, args.description)
        print(render(dataset, render_format))
        print("Dataset updated successfully.")

    def download(self, args, _render_format):
        """
        Download dataset.
        """
        # pylint: disable=no-self-use
        Dataset.download(args.dataset_id)



"""
Controller class for running commands (list, get, create, update)
on facility records.
"""
from __future__ import print_function

from mtclient.models.facility import Facility
from mtclient.models.instrument import Instrument
from mtclient.views import render

from .cli import ModelCliController


class FacilityController(ModelCliController):
    """
    Controller class for running commands (list, get, create, update)
    on facility records.
    """
    def __init__(self):
        super(FacilityController, self).__init__()
        self.allowed_commands = ["list", "get"]
        self.primary_key_arg = "facility_id"
        self.model = Facility

    def get(self, args, render_format):
        """
        Display facility record
        """
        super(FacilityController, self).get(args, render_format)
        if render_format == 'table':
            instruments = Instrument.list(
                filters="facility__id=%s" % args.facility_id)
            print(render(instruments, render_format))



"""
Controller class for running commands (list, get, create, update)
on experiment records.
"""
from __future__ import print_function

from mtclient.models.dataset import Dataset
from mtclient.models.experiment import Experiment
from mtclient.views import render

from .cli import ModelCliController


class ExperimentController(ModelCliController):
    """
    Controller class for running commands (list, get, create, update)
    on experiment records.
    """
    def __init__(self):
        super(ExperimentController, self).__init__()
        self.allowed_commands = ["list", "get", "create", "update"]
        self.primary_key_arg = "experiment_id"
        self.model = Experiment

    def list(self, args, render_format):
        """
        Display list of experiment records.
        """
        # pylint: disable=no-self-use
        experiments = Experiment.list(
            args.filter, args.limit, args.offset, args.order_by)
        print(render(experiments, render_format))

    def get(self, args, render_format):
        """
        Display experiment record.
        """
        # pylint: disable=no-self-use
        experiment = Experiment.objects.get(
            id=args.experiment_id, include_metadata=args.metadata)
        print(render(experiment, render_format))
        if render_format == 'table':
            datasets = Dataset.list(filters="experiments__id=%s" % args.experiment_id)
            if datasets:
                print(render(datasets, render_format, display_heading=False))

    def create(self, args, render_format):
        """
        Create experiment record.
        """
        # pylint: disable=no-self-use
        experiment = Experiment.create(
            args.title, args.description, args.institution, args.params)
        print(render(experiment, render_format))
        print("Experiment created successfully.")

    def update(self, args, render_format):
        """
        Update experiment record.
        """
        # pylint: disable=no-self-use
        experiment = Experiment.update(
            args.experiment_id, args.title, args.description)
        print(render(experiment, render_format))
        print("Experiment updated successfully.")



"""
test_schema_cli.py

Tests for querying the MyTardis REST API's schema endpoints
via the command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.schema import SchemaController


def test_schema_list_cli_json(capfd):
    """
    Test listing schemas, requesting output in JSON format
    """
    mock_schema_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "hidden": True,
                "immutable": True,
                "name": "Schema Name",
                "namespace": "http://schema/namespace/1",
                "resource_uri": "/api/v1/schema/1/",
                "subtype": None,
                "type": 1
            }
        ]
    }
    mock_schema_list_response = json.dumps(mock_schema_list)
    with requests_mock.Mocker() as mocker:
        schema_list_url = "%s/api/v1/schema/?format=json" % config.url
        mocker.get(schema_list_url, text=mock_schema_list_response)
        schema_controller = SchemaController()
        args = Namespace(
            model='schema', command='list', json=True, verbose=False,
            limit=None, offset=None, order_by=None)

        schema_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_schema_list

        schema_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_schema_list

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'schema', 'list', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_schema_list
        sys.argv = sys_argv


def test_schema_list_cli_table(capfd):
    """
    Test listing schema records, requesting output in ASCII table format
    """
    mock_schema_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "hidden": True,
                "immutable": True,
                "name": "Schema Name",
                "namespace": "http://schema/namespace/1",
                "resource_uri": "/api/v1/schema/1/",
                "subtype": None,
                "type": 1
            }
        ]
    }
    mock_schema_list_response = json.dumps(mock_schema_list)
    expected = textwrap.dedent("""
        Model: Schema
        Query: %s/api/v1/schema/?format=json
        Total Count: 1
        Limit: 20
        Offset: 0

        +----+-------------+---------------------------+-------------------+---------+-----------+--------+
        | ID |    Name     |         Namespace         |       Type        | Subtype | Immutable | Hidden |
        +====+=============+===========================+===================+=========+===========+========+
        |  1 | Schema Name | http://schema/namespace/1 | Experiment schema |         | True      | True   |
        +----+-------------+---------------------------+-------------------+---------+-----------+--------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        schema_list_url = "%s/api/v1/schema/?format=json" % config.url
        mocker.get(schema_list_url, text=mock_schema_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'schema', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_schema_get_cli_json(capfd):
    """
    Test looking up and displaying a schema via the command-line interface
    """
    mock_schema = {
        "id": 1,
        "hidden": True,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace/1",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": None,
        "type": 1
    }
    mock_schema_get_response = json.dumps(mock_schema)
    with requests_mock.Mocker() as mocker:
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_schema_get_response)
        schema_controller = SchemaController()
        args = Namespace(
            model='schema', command='get', schema_id=1, params=False,
            json=True, verbose=False)
        schema_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_schema


def test_schema_get_cli_table(capfd):
    """
    Test getting schema record via the command-line interface,
    requesting output in ASCII table format
    """
    mock_schema = {
        "id": 1,
        "hidden": True,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace/1",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": None,
        "type": 1
    }
    mock_schema_get_response = json.dumps(mock_schema)
    mock_pname_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "name": "param-name1",
                "full_name": "Parameter Name1",
                "choices": "",
                "comparison_type": 8,
                "data_type": 7,
                "immutable": False,
                "is_searchable": False,
                "order": 1,
                "resource_uri": "/api/v1/parametername/1/",
                "schema": "/api/v1/schema/1/",
                "units": ""
            }
        ]
    }
    mock_pnames_response = json.dumps(mock_pname_list)
    expected = textwrap.dedent("""
        +--------------+---------------------------+
        | Schema field |           Value           |
        +==============+===========================+
        | ID           | 1                         |
        +--------------+---------------------------+
        | Name         | Schema Name               |
        +--------------+---------------------------+
        | Namespace    | http://schema/namespace/1 |
        +--------------+---------------------------+
        | Type         | Experiment schema         |
        +--------------+---------------------------+
        | Subtype      | None                      |
        +--------------+---------------------------+
        | Immutable    | True                      |
        +--------------+---------------------------+
        | Hidden       | True                      |
        +--------------+---------------------------+

        +------------------+-----------------+-------------+-------------+-------+-----------+---------------+-------+---------+-----------------+
        | ParameterName ID |    Full Name    |    Name     |  Data Type  | Units | Immutable | Is Searchable | Order | Choices | Comparison Type |
        +==================+=================+=============+=============+=======+===========+===============+=======+=========+=================+
        |                1 | Parameter Name1 | param-name1 | Long String |       | False     | False         | 1     |         | Contains        |
        +------------------+-----------------+-------------+-------------+-------+-----------+---------------+-------+---------+-----------------+
    """)
    with requests_mock.Mocker() as mocker:
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_schema_get_response)
        pnames_url = "%s/api/v1/parametername/?format=json&schema__id=1" % config.url
        mocker.get(pnames_url, text=mock_pnames_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'schema', 'get', '1', '--params']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv



"""
Views for MyTardis schema records
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_schema(schema, render_format):
    """
    Render schema

    :param schema: The schema to be rendered.
    :type schema: :class:`mtclient.models.schema.Schema`
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if render_format == 'json':
        return render_schema_as_json(schema)
    return render_schema_as_table(schema)


def render_schema_as_json(schema, indent=2, sort_keys=True):
    """
    Returns JSON representation of schema.

    :param schema: The schema to be rendered.
    :type schema: :class:`mtclient.models.schema.Schema`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        schema.response_dict, indent=indent, sort_keys=sort_keys)


def render_schema_as_table(schema):
    """
    Returns ASCII table view of schema.

    :param schema: The schema to be rendered.
    :type schema: :class:`mtclient.models.schema.Schema`
    """
    schema_parameter_names = ""

    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Schema field", "Value"])
    table.add_row(["ID", schema.id])
    table.add_row(["Name", schema.name])
    table.add_row(["Namespace", schema.namespace])
    table.add_row(["Type", schema.type])
    table.add_row(["Subtype", schema.subtype])
    table.add_row(["Immutable", str(bool(schema.immutable))])
    table.add_row(["Hidden", str(bool(schema.hidden))])
    schema_parameter_names += table.draw() + "\n"

    if not schema.parameter_names:
        return schema_parameter_names

    schema_parameter_names += "\n"
    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l', 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm', 'm', 'm', 'm'])
    table.header(["ParameterName ID", "Full Name", "Name", "Data Type",
                  "Units", "Immutable", "Is Searchable", "Order", "Choices",
                  "Comparison Type"])
    for parameter_name in schema.parameter_names:
        table.add_row([parameter_name.id,
                       parameter_name.full_name.encode('utf8', 'ignore'),
                       parameter_name.name, parameter_name.data_type,
                       parameter_name.units.encode('utf8', 'ignore'),
                       str(bool(parameter_name.immutable)),
                       str(bool(parameter_name.is_searchable)),
                       parameter_name.order,
                       parameter_name.choices,
                       parameter_name.comparison_type])
    schema_parameter_names += table.draw() + "\n"

    return schema_parameter_names


def render_schemas(schemas, render_format, display_heading=True):
    """
    Render schemas

    :param schemas: The `ResultSet` of schemas to be rendered.
    :type schemas: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_schemas_as_json(schemas)
    return render_schemas_as_table(schemas, display_heading)


def render_schemas_as_json(schemas, indent=2, sort_keys=True):
    """
    Returns JSON representation of schemas.

    :param schemas: The result set of schemas boxes to be displayed.
    :type schemas: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        schemas.response_dict, indent=indent, sort_keys=sort_keys)


def render_schemas_as_table(schemas, display_heading=True):
    """
    Returns ASCII table view of schemas.

    :param schemas: The schemas to be rendered.
    :type schemas: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: Schema\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (schemas.url, schemas.total_count,
           schemas.limit, schemas.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
    table.header(["ID", "Name", "Namespace", "Type", "Subtype", "Immutable",
                  "Hidden"])
    for schema in schemas:
        table.add_row([schema.id, schema.name, schema.namespace,
                       schema.type, schema.subtype or '',
                       str(bool(schema.immutable)), str(bool(schema.hidden))])
    return heading + table.draw() + "\n"



"""
Controller class for running commands (list, get, create, update)
on instrument records.
"""
from __future__ import print_function

from mtclient.models.instrument import Instrument
from mtclient.views import render

from .cli import ModelCliController


class InstrumentController(ModelCliController):
    """
    Controller class for running commands (list, get, create, update)
    on instrument records.
    """
    def __init__(self):
        super(InstrumentController, self).__init__()
        self.allowed_commands = ["list", "get", "create", "update"]
        self.primary_key_arg = "instrument_id"
        self.model = Instrument

    def list(self, args, render_format):
        """
        Display list of instrument records.
        """
        # pylint: disable=no-self-use
        facility_id = getattr(args, "facility_id", getattr(args, "facility", None))
        filters = ""
        if facility_id:
            filters = "facility__id=%s" % facility_id
        instruments = Instrument.list(
            filters, args.limit, args.offset, args.order_by)
        print(render(instruments, render_format))

    def get(self, args, render_format):
        """
        Display instrument record.
        """
        # pylint: disable=no-self-use
        instrument = Instrument.objects.get(id=args.instrument_id)
        print(render(instrument, render_format))

    def create(self, args, render_format):
        """
        Create instrument record.
        """
        # pylint: disable=no-self-use
        facility_id = getattr(args, "facility_id", getattr(args, "facility", None))
        instrument = Instrument.create(facility_id, args.name)
        print(render(instrument, render_format))
        print("Instrument created successfully.")

    def update(self, args, render_format):
        """
        Update instrument record.
        """
        # pylint: disable=no-self-use
        instrument = Instrument.update(args.instrument_id, args.name)
        print(render(instrument, render_format))
        print("Instrument updated successfully.")



"""
Views for MyTardis records.
"""
from __future__ import print_function

import json
from texttable import Texttable

from ..utils import human_readable_size_string


def render_datafile(datafile, render_format):
    """
    Render datafile

    :param datafile: The datafile to be rendered.
    :type datafile: :class:`mtclient.models.datafile.DataFile`
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if render_format == 'json':
        return render_datafile_as_json(datafile)
    return render_datafile_as_table(datafile)


def render_datafile_as_json(datafile, indent=2, sort_keys=True):
    """
    Returns JSON representation of datafile.

    :param datafile: The datafile to be rendered.
    :type datafile: :class:`mtclient.models.datafile.DataFile`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        datafile.response_dict, indent=indent, sort_keys=sort_keys)


def render_datafile_as_table(datafile):
    """
    Returns ASCII table view of datafile.

    :param datafile: The datafile to be rendered.
    :type datafile: :class:`mtclient.models.datafile.DataFile`
    """
    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["DataFile field", "Value"])
    table.add_row(["ID", datafile.id])
    table.add_row(["Dataset", datafile.dataset])
    locations = [replica.location for replica in datafile.replicas]
    table.add_row(["Storage Box", "\n".join(locations)])
    table.add_row(["Directory", datafile.directory])
    table.add_row(["Filename", datafile.filename])
    uris = [replica.uri for replica in datafile.replicas]
    table.add_row(["URI", "\n".join(uris)])
    table.add_row(["Verified", str(datafile.verified)])
    table.add_row(["Size", human_readable_size_string(datafile.size)])
    table.add_row(["MD5 Sum", datafile.md5sum])
    datafile_and_param_sets = table.draw() + "\n"

    for datafile_param_set in datafile.parameter_sets:
        datafile_and_param_sets += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
        table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
        table.header(["DataFileParameter ID", "Schema", "Parameter Name",
                      "String Value", "Numerical Value", "Datetime Value",
                      "Link ID"])
        for datafile_param in datafile_param_set.parameters:
            table.add_row([datafile_param.id,
                           datafile_param.name.schema.name,
                           datafile_param.name.name,
                           datafile_param.string_value or "",
                           datafile_param.numerical_value or '',
                           datafile_param.datetime_value or '',
                           datafile_param.link_id or ''])
        datafile_and_param_sets += table.draw() + "\n"

    return datafile_and_param_sets


def render_datafiles(datafiles, render_format, display_heading=True):
    """
    Render datafiles

    :param datafiles: The `ResultSet` of datafiles to be rendered.
    :type datafiles: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_datafiles_as_json(datafiles)
    return render_datafiles_as_table(datafiles, display_heading)


def render_datafiles_as_json(datafiles, indent=2, sort_keys=True):
    """
    Returns JSON representation of datafiles.

    :param datafiles: The result set of datafiles to be displayed.
    :type datafiles: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        datafiles.response_dict, indent=indent, sort_keys=sort_keys)


def render_datafiles_as_table(datafiles, display_heading=True):
    """
    Returns ASCII table view of datafiles.

    :param datafiles: The datafiles to be rendered.
    :type datafiles: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: DataFile\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (datafiles.url, datafiles.total_count, datafiles.limit,
           datafiles.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
    table.header(["DataFile ID", "Filename", "Storage Box",
                  "URI", "Verified", "Size", "MD5 Sum"])
    for datafile in datafiles:
        uris = [replica.uri for replica in datafile.replicas]
        locations = [replica.location for replica in datafile.replicas]
        table.add_row([datafile.id, datafile.filename, "\n".join(locations),
                       "\n".join(uris), str(datafile.verified),
                       human_readable_size_string(datafile.size),
                       datafile.md5sum])
    return heading + table.draw() + "\n"



"""
argparser/experiment.py
"""
import textwrap


def build_experiment_parser(argument_parser):
    """
    Builds parsing rules for experiment-related command-line interface arguments.
    """
    # pylint: disable=too-many-locals
    experiment_help = \
        "Display a list of experiment records or a single experiment record."
    experiment_usage = "mytardis experiment [-h] {list,get,create,update} ..."
    experiment_parser = \
        argument_parser.model_parsers.add_parser("experiment",
                                                 help=experiment_help,
                                                 usage=experiment_usage)
    experiment_command_parsers = \
        experiment_parser.add_subparsers(help='available commands',
                                         dest='command')

    experiment_list_help = "Display a list of experiment records."
    experiment_list_usage = textwrap.dedent("""\
        mytardis experiment list
            [--limit LIMIT] [--offset OFFSET] [--order_by ORDER_BY] [--json]
            [--filter FILTER]

          EXAMPLE

          $ mytardis experiment list

          Model: Experiment
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/experiment/?format=json
          Total Count: 4
          Limit: 20
          Offset: 0

          +----+-------------------+-----------------------------------+
          | ID |    Institution    |               Title               |
          +====+===================+===================================+
          | 20 | Monash University | James Exp 001                     |
          +----+-------------------+-----------------------------------+
          | 13 | Monash University | A's Test Instrument - Test User1  |
          +----+-------------------+-----------------------------------+
          | 22 | Monash University | James Test Exp 003                |
          +----+-------------------+-----------------------------------+
          | 14 | Monash University | A's Test Instrument - Test User2  |
          +----+-------------------+-----------------------------------+
        """)
    experiment_command_list_parser = \
        experiment_command_parsers.add_parser("list",
                                              help=experiment_list_help,
                                              usage=experiment_list_usage)
    experiment_command_list_parser.add_argument(
        "--limit", help="Maximum number of results to return.")
    experiment_command_list_parser.add_argument(
        "--offset",
        help="Skip this many records from the start of the result set.")
    experiment_command_list_parser.add_argument(
        "--order_by", default="-created_time",
        help="Order by this field.")
    experiment_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")
    experiment_command_list_parser.add_argument(
        "--filter", help="Filter on these fields, e.g. \"title=Exp Title\".")

    experiment_get_help = "Display a single experiment record."
    experiment_get_usage = textwrap.dedent("""\
        mytardis experiment get [-h] [--json] experiment_id

          EXAMPLE

          $ mytardis experiment get 20

          +------------------+-------------------+
          | Experiment field |       Value       |
          +==================+===================+
          | ID               | 20                |
          +------------------+-------------------+
          | Institution      | Monash University |
          +------------------+-------------------+
          | Title            | James Exp 001     |
          +------------------+-------------------+
          | Description      |                   |
          +------------------+-------------------+

          +------------+------------------------+------------------------+------------+
          | Dataset ID |     Experiment(s)      |      Description       | Instrument |
          +============+========================+========================+============+
          |         34 | /api/v1/experiment/20/ | James Test Dataset 001 | None       |
          +------------+------------------------+------------------------+------------+
          |         33 | /api/v1/experiment/20/ | James Test Dataset 003 | None       |
          +------------+------------------------+------------------------+------------+
          |         32 | /api/v1/experiment/20/ | James Test Dataset 002 | None       |
          +------------+------------------------+------------------------+------------+
          |         31 | /api/v1/experiment/20/ | James Test Dataset 001 | None       |
          +------------+------------------------+------------------------+------------+
        """)
    experiment_command_get_parser = \
        experiment_command_parsers.add_parser("get",
                                              help=experiment_get_help,
                                              usage=experiment_get_usage)
    experiment_command_get_parser.add_argument("experiment_id",
                                               help="The experiment ID.")
    experiment_command_get_parser.add_argument(
        "--metadata", action='store_true', help="Include metadata.")
    experiment_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    experiment_create_help = "Create an experiment record."
    experiment_create_usage = textwrap.dedent("""\
        mytardis experiment create
            [--description DESCRIPTION] [--institution INSTITUTION] [--params PARAMS] title

          EXAMPLE

          First let's look up an experiment schema we can use to add some
          metadata to the new experiment record.

          $ mytardis schema list

          +----+---------------------------+-----------------------------------------------------------+-------------------+---------+-----------+--------+
          | ID |           Name            |                         Namespace                         |       Type        | Subtype | Immutable | Hidden |
          +====+===========================+===========================================================+===================+=========+===========+========+
           ...  ...                         ...                                                         ...                 ...       ...         ...
          +----+---------------------------+-----------------------------------------------------------+-------------------+---------+-----------+--------+
          | 12 | Sample Experiment Schema  | https://mytardis.org/schemas/sample-experiment-schema     | Experiment schema |         | False     | False  |
          +----+---------------------------+-----------------------------------------------------------+-------------------+---------+-----------+--------+

          "mytardis schema get <schema_id>" displays the parameters
          associated with this schema:

          $ mytardis schema get 12
          +--------------+-------------------------------------------------------+
          | Schema field |                         Value                         |
          +==============+=======================================================+
          | ID           | 12                                                    |
          +--------------+-------------------------------------------------------+
          | Name         | Sample Experiment Schema                              |
          +--------------+-------------------------------------------------------+
          | Namespace    | https://mytardis.org/schemas/sample-experiment-schema |
          +--------------+-------------------------------------------------------+
          | Type         | Experiment schema                                     |
          +--------------+-------------------------------------------------------+
          | Subtype      |                                                       |
          +--------------+-------------------------------------------------------+
          | Immutable    | False                                                 |
          +--------------+-------------------------------------------------------+
          | Hidden       | False                                                 |
          +--------------+-------------------------------------------------------+

          +------------------+-----------------------+-----------------------+-----------+-------+-----------+---------------+-------+---------+-----------------+
          | ParameterName ID |       Full Name       |         Name          | Data Type | Units | Immutable | Is Searchable | Order | Choices | Comparison Type |
          +==================+=======================+=======================+===========+=======+===========+===============+=======+=========+=================+
          |               33 | Sample Parameter Name | sample_parameter_name | String    |       | False     | False         | 9999  |         | Exact value     |
          +------------------+-----------------------+-----------------------+-----------+-------+-----------+---------------+-------+---------+-----------------+

          Now that we know the schema's namespace, the parameter name(s) and
          their data type(s), we can create some experiment metadata.

          Below we use cat to create the metadata file, but you can use
          any text editor.

          $ cat << EOF > params.json
          [
              {
                  "schema": "https://mytardis.org/schemas/sample-experiment-schema",
                  "parameters": [
                      {
                          "name": "sample_parameter_name",
                          "value": "Sample Parameter Value"
                      }
                  ]
              }
          ]
          EOF

          Now we create the experiment record, using the metadata file
          we just created:

          $ mytardis experiment create --params params.json "Exp With Params"
          +------------------+-------------------+
          | Experiment field |       Value       |
          +==================+===================+
          | ID               | 28                |
          +------------------+-------------------+
          | Institution      | Monash University |
          +------------------+-------------------+
          | Title            | Exp With Params   |
          +------------------+-------------------+
          | Description      |                   |
          +------------------+-------------------+

          +------------------------+--------------------------+-----------------------+------------------------+-----------------+----------------+---------+
          | ExperimentParameter ID |          Schema          |    Parameter Name     |      String Value      | Numerical Value | Datetime Value | Link ID |
          +========================+==========================+=======================+========================+=================+================+=========+
          |                     35 | Sample Experiment Schema | Sample Parameter Name | Sample Parameter Value |                 |                |         |
          +------------------------+--------------------------+-----------------------+------------------------+-----------------+----------------+---------+

          Experiment created successfully.
        """)
    experiment_cmd_create_parser = \
        experiment_command_parsers.add_parser("create",
                                              help=experiment_create_help,
                                              usage=experiment_create_usage)
    experiment_cmd_create_parser.add_argument(
        "title", help="The experiment title to create.")
    experiment_cmd_create_parser.add_argument(
        "--description", help="A description of the experiment.")
    experiment_cmd_create_parser.add_argument(
        "--institution", help="The institution of the experiment.")
    experiment_cmd_create_parser.add_argument(
        "--params", help="A JSON file containing experiment parameters.")

    experiment_update_help = "Update an experiment record."
    experiment_update_usage = textwrap.dedent("""\
        mytardis experiment update
            [--title TITLE] [--description DESCRIPTION] experiment_id

          EXAMPLE
          $ mytardis experiment update --title "Renamed Exp" 20
          +------------------+-------------------+
          | Experiment field |       Value       |
          +==================+===================+
          | ID               | 20                |
          +------------------+-------------------+
          | Institution      | Monash University |
          +------------------+-------------------+
          | Title            | Renamed Exp       |
          +------------------+-------------------+
          | Description      |                   |
          +------------------+-------------------+

          Experiment updated successfully.
        """)
    experiment_cmd_update_parser = \
        experiment_command_parsers.add_parser("update",
                                              help=experiment_update_help,
                                              usage=experiment_update_usage)
    experiment_cmd_update_parser.add_argument(
        "experiment_id", help="The ID of the experiment to update.")
    experiment_cmd_update_parser.add_argument(
        "--title", help="The new title of the experiment.")
    experiment_cmd_update_parser.add_argument(
        "--description", help="The new description of the experiment.")



"""
Views for MyTardis experiment records
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_experiment(experiment, render_format):
    """
    Render experiment

    :param experiment: The experiment to be rendered.
    :type experiment: :class:`mtclient.models.experiment.Experiment`
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if render_format == 'json':
        return render_experiment_as_json(experiment)
    return render_experiment_as_table(experiment)


def render_experiment_as_json(experiment, indent=2, sort_keys=True):
    """
    Returns JSON representation of experiment.

    :param experiment: The experiment to be rendered.
    :type experiment: :class:`mtclient.models.experiment.Experiment`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        experiment.response_dict, indent=indent, sort_keys=sort_keys)


def render_experiment_as_table(experiment):
    """
    Returns ASCII table view of experiment.

    :param experiment: The experiment to be rendered.
    :type experiment: :class:`mtclient.models.experiment.Experiment`
    """
    exp_and_param_sets = ""

    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Experiment field", "Value"])
    table.add_row(["ID", experiment.id])
    table.add_row(["Institution", experiment.institution_name])
    table.add_row(["Title", experiment.title])
    table.add_row(["Description", experiment.description])
    exp_and_param_sets += table.draw() + "\n"

    for exp_param_set in experiment.parameter_sets:
        exp_and_param_sets += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
        table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
        table.header(["ExperimentParameter ID", "Schema", "Parameter Name",
                      "String Value", "Numerical Value", "Datetime Value",
                      "Link ID"])
        for exp_param in exp_param_set.parameters:
            table.add_row([exp_param.id,
                           exp_param.name.schema.name,
                           exp_param.name.name,
                           exp_param.string_value or '',
                           exp_param.numerical_value or '',
                           exp_param.datetime_value or '',
                           exp_param.link_id or ''])
        exp_and_param_sets += table.draw() + "\n"

    return exp_and_param_sets


def render_experiments(experiments, render_format, display_heading=True):
    """
    Render experiments

    :param experiments: The `ResultSet` of experiments to be rendered.
    :type experiments: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_experiments_as_json(experiments)
    return render_experiments_as_table(experiments, display_heading)


def render_experiments_as_json(experiments, indent=2, sort_keys=True):
    """
    Returns JSON representation of experiments.

    :param experiments: The result set of experiments to be displayed.
    :type experiments: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        experiments.response_dict, indent=indent, sort_keys=sort_keys)


def render_experiments_as_table(experiments, display_heading=True):
    """
    Returns ASCII table view of experiments.

    :param experiments: The experiments to be rendered.
    :type experiments: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: Experiment\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (experiments.url, experiments.total_count,
           experiments.limit, experiments.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Institution", "Title"])
    for experiment in experiments:
        table.add_row([experiment.id, experiment.institution_name,
                       experiment.title])
    return heading + table.draw() + "\n"



"""
test_dataset_cli.py

Tests for querying the MyTardis REST API's dataset endpoints
via the command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.dataset import DatasetController


def test_dataset_list_cli_json(capfd):
    """
    Test listing datasets, requesting output in JSON format
    """
    mock_dataset_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "description": "dataset description",
                "experiments": [
                    "/api/v1/experiment/1/"
                ],
                "immutable": False,
                "instrument": None,
                "parameter_sets": [],
                "resource_uri": "/api/v1/dataset/1/"
            }
        ]
    }
    mock_ds_list_response = json.dumps(mock_dataset_list)
    with requests_mock.Mocker() as mocker:
        dataset_list_url = "%s/api/v1/dataset/?format=json&experiments__id=1" % config.url
        mocker.get(dataset_list_url, text=mock_ds_list_response)
        ds_controller = DatasetController()
        args = Namespace(
            model='dataset', command='list', exp='1', json=True, verbose=False,
            filter=None, limit=None, offset=None, order_by=None)

        ds_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_dataset_list

        ds_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_dataset_list

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'dataset', 'list', '--exp', '1', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_dataset_list
        sys.argv = sys_argv


def test_dataset_list_cli_table(capfd):
    """
    Test listing dataset records, requesting output in ASCII table format
    """
    mock_dataset_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "description": "dataset description",
                "experiments": [
                    "/api/v1/experiment/1/"
                ],
                "immutable": False,
                "instrument": None,
                "parameter_sets": [],
                "resource_uri": "/api/v1/dataset/1/"
            }
        ]
    }
    mock_ds_list_response = json.dumps(mock_dataset_list)
    expected = textwrap.dedent("""
        Model: Dataset
        Query: %s/api/v1/dataset/?format=json
        Total Count: 1
        Limit: 20
        Offset: 0
        
        +------------+-----------------------+---------------------+------------+
        | Dataset ID |     Experiment(s)     |     Description     | Instrument |
        +============+=======================+=====================+============+
        |          1 | /api/v1/experiment/1/ | dataset description | None       |
        +------------+-----------------------+---------------------+------------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        dataset_list_url = "%s/api/v1/dataset/?format=json" % config.url
        mocker.get(dataset_list_url, text=mock_ds_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'dataset', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_dataset_get_cli_json(capfd):
    """
    Test looking up and displaying an dataset via the command-line interface
    """
    mock_dataset = {
        "id": 1,
        "description": "dataset description",
        "experiments": [
            "/api/v1/experiment/1/"
        ],
        "immutable": False,
        "instrument": None,
        "parameter_sets": [],
        "resource_uri": "/api/v1/dataset/1/"
    }
    mock_dataset_get_response = json.dumps(mock_dataset)
    with requests_mock.Mocker() as mocker:
        get_dataset_url = "%s/api/v1/dataset/1/?format=json" % config.url
        mocker.get(get_dataset_url, text=mock_dataset_get_response)
        ds_controller = DatasetController()
        args = Namespace(
            model='dataset', command='get', dataset_id=1, json=True,
            verbose=False, metadata=False)
        ds_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_dataset


def test_dataset_get_cli_table(capfd):
    """
    Test getting dataset record via the command-line interface,
    requesting output in ASCII table format
    """
    mock_dataset = {
        "id": 1,
        "description": "dataset description",
        "experiments": [
            "/api/v1/experiment/1/"
        ],
        "immutable": False,
        "instrument": None,
        "parameter_sets": [
            {
                "id": 1,
                "dataset": "/api/v1/dataset/1/",
                "schema": {
                    "id": 1,
                    "name": "Series Metadata",
                    "namespace": "http://schema/namespace",
                    "hidden": False,
                    "immutable": True,
                    "subtype": "",
                    "type": 1
                },
                "parameters": [
                    {
                        "id": 1,
                        "name": "/api/v1/parametername/1/",
                        "parameterset": "/api/v1/datasetparameterset/1/",
                        "resource_uri": "/api/v1/datasetparameter/1/",
                        "string_value": "param value",
                        "numerical_value": None,
                        "datetime_value": None,
                        "link_id": None,
                        "value": None
                    }
                ]
            }
        ],
        "resource_uri": "/api/v1/dataset/1/"
    }
    mock_dataset_get_response = json.dumps(mock_dataset)
    mock_schema = {
        "hidden": False,
        "id": 1,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": "",
        "type": 1
    }
    mock_schema_response = json.dumps(mock_schema)
    mock_pname = {
        "id": 1,
        "choices": "",
        "comparison_type": 1,
        "data_type": 2,
        "full_name": "Parameter Name",
        "immutable": True,
        "is_searchable": False,
        "name": "param name",
        "order": 1,
        "resource_uri": "/api/v1/parametername/1/",
        "schema": "/api/v1/schema/1/",
        "units": ""
    }
    mock_pname_response = json.dumps(mock_pname)
    mock_datafile_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 0
        },
        "objects": [
        ]
    }
    mock_df_list_response = json.dumps(mock_datafile_list)
    expected = textwrap.dedent("""
        +---------------+-----------------------+
        | Dataset field |         Value         |
        +===============+=======================+
        | ID            | 1                     |
        +---------------+-----------------------+
        | Experiment(s) | /api/v1/experiment/1/ |
        +---------------+-----------------------+
        | Description   | dataset description   |
        +---------------+-----------------------+
        | Instrument    | None                  |
        +---------------+-----------------------+

        +---------------------+-------------+----------------+--------------+-----------------+----------------+---------+
        | DatasetParameter ID |   Schema    | Parameter Name | String Value | Numerical Value | Datetime Value | Link ID |
        +=====================+=============+================+==============+=================+================+=========+
        |                   1 | Schema Name | param name     | param value  |                 |                |         |
        +---------------------+-------------+----------------+--------------+-----------------+----------------+---------+
    """)
    with requests_mock.Mocker() as mocker:
        get_dataset_url = "%s/api/v1/dataset/1/?format=json" % config.url
        mocker.get(get_dataset_url, text=mock_dataset_get_response)
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_schema_response)
        get_pname_url = "%s/api/v1/parametername/1/?format=json" % config.url
        mocker.get(get_pname_url, text=mock_pname_response)
        df_list_url = "%s/api/v1/dataset_file/?format=json&dataset__id=1" % config.url
        mocker.get(df_list_url, text=mock_df_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'dataset', 'get', '1', '--metadata']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv



"""
test_storagebox_cli.py

Tests for querying the MyTardis REST API's storage box endpoints
via the command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.storagebox import StorageBoxController


def test_storagebox_list_cli_json(capfd):
    """
    Test listing storage boxes, requesting output in JSON format
    """
    mock_boxes_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "django_storage_class": "tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage",
                "name": "local box at /home/mytardis/var/local",
                "description": "Storage box description",
                "status": "online",
                "max_size": None,
                "attributes": [
                ],
                "options": [
                    {
                        "id": 1,
                        "key": "location",
                        "resource_uri": "/api/v1/storageboxoption/1/",
                        "storage_box": "/api/v1/storagebox/1/",
                        "value": "/home/mytardis/var/local",
                        "value_type": "string"
                    }
                ],
                "resource_uri": "/api/v1/storagebox/1/"
            }
        ]
    }
    mock_boxes_list_response = json.dumps(mock_boxes_list)
    with requests_mock.Mocker() as mocker:
        boxes_list_url = "%s/api/v1/storagebox/?format=json" % config.url
        mocker.get(boxes_list_url, text=mock_boxes_list_response)
        storagebox_controller = StorageBoxController()
        args = Namespace(
            model='storagebox', command='list', json=True, verbose=False,
            limit=None, offset=None, order_by=None)

        storagebox_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_boxes_list

        storagebox_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_boxes_list

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'storagebox', 'list', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_boxes_list
        sys.argv = sys_argv


def test_storagebox_list_cli_table(capfd):
    """
    Test listing storage box records, requesting output in ASCII table format
    """
    mock_boxes_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "django_storage_class": "tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage",
                "name": "local box at /home/mytardis/var/local",
                "description": "Storage box description",
                "status": "online",
                "max_size": None,
                "attributes": [
                ],
                "options": [
                    {
                        "id": 1,
                        "key": "location",
                        "resource_uri": "/api/v1/storageboxoption/1/",
                        "storage_box": "/api/v1/storagebox/1/",
                        "value": "/home/mytardis/var/local",
                        "value_type": "string"
                    }
                ],
                "resource_uri": "/api/v1/storagebox/1/",
            }
        ]
    }
    mock_boxes_list_response = json.dumps(mock_boxes_list)
    expected = textwrap.dedent("""
        Model: StorageBox
        Query: %s/api/v1/storagebox/?format=json
        Total Count: 1
        Limit: 20
        Offset: 0
              
        +----+---------------------------------------+-------------------------+
        | ID |                 Name                  |       Description       |
        +====+=======================================+=========================+
        |  1 | local box at /home/mytardis/var/local | Storage box description |
        +----+---------------------------------------+-------------------------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        boxes_list_url = "%s/api/v1/storagebox/?format=json" % config.url
        mocker.get(boxes_list_url, text=mock_boxes_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'storagebox', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_storagebox_get_cli_json(capfd):
    """
    Test looking up and displaying a storage box via the command-line interface
    """
    mock_storagebox = {
        "id": 1,
        "django_storage_class": "tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage",
        "name": "local box at /home/mytardis/var/local",
        "description": "Storage box description",
        "status": "online",
        "max_size": None,
        "attributes": [
        ],
        "options": [
            {
                "id": 1,
                "key": "location",
                "resource_uri": "/api/v1/storageboxoption/1/",
                "storage_box": "/api/v1/storagebox/1/",
                "value": "/home/mytardis/var/local",
                "value_type": "string"
            }
        ],
        "resource_uri": "/api/v1/storagebox/1/",
    }
    mock_storagebox_get_response = json.dumps(mock_storagebox)
    with requests_mock.Mocker() as mocker:
        get_storagebox_url = "%s/api/v1/storagebox/1/?format=json" % config.url
        mocker.get(get_storagebox_url, text=mock_storagebox_get_response)
        storagebox_controller = StorageBoxController()
        args = Namespace(
            model='storagebox', command='get', storage_box_id=1, params=False,
            json=True, verbose=False)
        storagebox_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_storagebox


def test_storagebox_get_cli_table(capfd):
    """
    Test getting storage box record via the command-line interface,
    requesting output in ASCII table format
    """
    mock_storagebox = {
        "id": 1,
        "django_storage_class": "tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage",
        "name": "local box at /home/mytardis/var/local",
        "description": "Storage box description",
        "status": "online",
        "max_size": None,
        "attributes": [
            {
                "id": 1,
                "key": "test_attr_key",
                "resource_uri": "/api/v1/storageboxattribute/1/",
                "storage_box": "/api/v1/storagebox/1/",
                "value": "test_attr_value",
                "value_type": "string"
            }
        ],
        "options": [
            {
                "id": 1,
                "key": "location",
                "resource_uri": "/api/v1/storageboxoption/1/",
                "storage_box": "/api/v1/storagebox/1/",
                "value": "/home/mytardis/var/local",
                "value_type": "string"
            }
        ],
        "resource_uri": "/api/v1/storagebox/1/",
    }
    mock_storagebox_get_response = json.dumps(mock_storagebox)
    expected = textwrap.dedent("""
        +----------------------+-------------------------------------------------------------+
        |   StorageBox field   |                            Value                            |
        +======================+=============================================================+
        | ID                   | 1                                                           |
        +----------------------+-------------------------------------------------------------+
        | Name                 | local box at /home/mytardis/var/local                       |
        +----------------------+-------------------------------------------------------------+
        | Description          | Storage box description                                     |
        +----------------------+-------------------------------------------------------------+
        | Django Storage Class | tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage |
        +----------------------+-------------------------------------------------------------+
        | Max Size             | None                                                        |
        +----------------------+-------------------------------------------------------------+
        | Status               | online                                                      |
        +----------------------+-------------------------------------------------------------+

        +----------------------+--------------------------+
        | StorageBoxOption Key |  StorageBoxOption Value  |
        +======================+==========================+
        |             location | /home/mytardis/var/local |
        +----------------------+--------------------------+

        +-------------------------+---------------------------+
        | StorageBoxAttribute Key | StorageBoxAttribute Value |
        +=========================+===========================+
        |           test_attr_key | test_attr_value           |
        +-------------------------+---------------------------+
    """)
    with requests_mock.Mocker() as mocker:
        get_storagebox_url = "%s/api/v1/storagebox/1/?format=json" % config.url
        mocker.get(get_storagebox_url, text=mock_storagebox_get_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'storagebox', 'get', '1']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv



"""
tests/config/test_config.py
"""
import json
import os
import sys
import tempfile
import textwrap

import pytest

from mtclient.conf import config
from mtclient.utils.exceptions import InvalidConfig


def test_validation():
    """
    Test configuration validation
    """
    assert config.validate()

    config_url = config.url
    config.url = ""
    with pytest.raises(InvalidConfig) as excinfo:
        config.validate()
    config.url = config_url
    assert excinfo.type == InvalidConfig
    assert "MyTardis URL is missing from config" in str(excinfo.value)

    config_url = config.url
    config.url = "invalid://example.com"
    with pytest.raises(InvalidConfig) as excinfo:
        config.validate()
    config.url = config_url
    assert excinfo.type == InvalidConfig
    assert "Invalid MyTardis URL found in config" in str(excinfo.value)

    config_username = config.username
    config.username = ""
    with pytest.raises(InvalidConfig) as excinfo:
        config.validate()
    config.username = config_username
    assert excinfo.type == InvalidConfig
    assert "MyTardis username is missing from config" in str(excinfo.value)

    config_apikey = config.apikey
    config.apikey = ""
    with pytest.raises(InvalidConfig) as excinfo:
        config.validate()
    config.apikey = config_apikey
    assert excinfo.type == InvalidConfig
    assert "MyTardis API key is missing from config" in str(excinfo.value)

def test_string_repr():
    """
    Test string representation
    """
    config_dict = config.__dict__
    # datasets_path is a property method, whose value
    # gets displayed in the string represetnation:
    config_dict["datasets_path"] = config.datasets_path
    assert json.loads(str(config)) == config_dict

def test_save():
    """
    Test saving config to disk
    """
    with tempfile.NamedTemporaryFile() as tmpfile:
        tmpfile_path = tmpfile.name

    config.save(tmpfile_path)

    with open(tmpfile_path, 'r') as config_file:
        config_file_content = config_file.read()

    expected = textwrap.dedent("""
    [mytardisclient]
    url = %s
    username = %s
    apikey = %s
    """ % (config.url, config.username, config.apikey))

    assert config_file_content.strip() == expected.strip()
    try:
        os.remove(tmpfile_path)
    except IOError as err:
        sys.stderr.write("%s\n" % err)



"""
test_datafile_cli.py

Tests for querying the MyTardis REST API's datafile endpoints
via the command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.datafile import DataFileController


def test_datafile_list_cli_json(capfd):
    """
    Test listing datafiles, requesting output in JSON format
    """
    mock_datafile_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "created_time": "2016-11-10T13:50:25.258483",
                "dataset": "/api/v1/dataset/1/",
                "directory": "",
                "filename": "testfile1.txt",
                "md5sum": "bogus",
                "mimetype": "text/plain",
                "modification_time": None,
                "parameter_sets": [
                ],
                "replicas": [
                    {
                        "datafile": "/api/v1/dataset_file/1/",
                        "id": 1,
                        "location": "local box at /home/mytardis/var/local",
                        "resource_uri": "/api/v1/replica/1/",
                        "uri": "subdir/testfile1.txt",
                        "verified": True
                    }
                ],
                "resource_uri": "/api/v1/dataset_file/1/",
                "size": 32
            }
        ]
    }
    mock_df_list_response = json.dumps(mock_datafile_list)
    with requests_mock.Mocker() as mocker:
        datafile_list_url = "%s/api/v1/dataset_file/?format=json&dataset__id=1" % config.url
        mocker.get(datafile_list_url, text=mock_df_list_response)
        df_controller = DataFileController()
        args = Namespace(
            model='datafile', command='list', dataset='1',
            directory=None, filename=None,
            json=True, verbose=False,
            filter=None, limit=None, offset=None, order_by=None)

        df_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_datafile_list

        df_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_datafile_list

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'datafile', 'list', '--dataset', '1', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_datafile_list
        sys.argv = sys_argv


def test_datafile_list_cli_table(capfd):
    """
    Test listing datafile records, requesting output in ASCII table format
    """
    mock_datafile_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "id": 1,
                "created_time": "2016-11-10T13:50:25.258483",
                "dataset": "/api/v1/dataset/1/",
                "directory": "",
                "filename": "testfile1.txt",
                "md5sum": "bogus",
                "mimetype": "text/plain",
                "modification_time": None,
                "parameter_sets": [
                ],
                "replicas": [
                    {
                        "datafile": "/api/v1/dataset_file/1/",
                        "id": 1,
                        "location": "local box at /home/mytardis/var/local",
                        "resource_uri": "/api/v1/replica/1/",
                        "uri": "subdir/testfile1.txt",
                        "verified": True
                    }
                ],
                "resource_uri": "/api/v1/dataset_file/1/",
                "size": 32
            }
        ]
    }
    mock_df_list_response = json.dumps(mock_datafile_list)
    expected = textwrap.dedent("""
        Model: DataFile
        Query: %s/api/v1/dataset_file/?format=json
        Total Count: 1
        Limit: 20
        Offset: 0

        +-------------+---------------+---------------------------------------+----------------------+----------+-----------+---------+
        | DataFile ID |   Filename    |              Storage Box              |         URI          | Verified |   Size    | MD5 Sum |
        +=============+===============+=======================================+======================+==========+===========+=========+
        |           1 | testfile1.txt | local box at /home/mytardis/var/local | subdir/testfile1.txt | True     |  32 bytes | bogus   |
        +-------------+---------------+---------------------------------------+----------------------+----------+-----------+---------+
    """) % config.url
    with requests_mock.Mocker() as mocker:
        datafile_list_url = "%s/api/v1/dataset_file/?format=json" % config.url
        mocker.get(datafile_list_url, text=mock_df_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'datafile', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_datafile_get_cli_json(capfd):
    """
    Test looking up and displaying an datafile via the command-line interface
    """
    mock_datafile = {
        "id": 1,
        "created_time": "2016-11-10T13:50:25.258483",
        "dataset": "/api/v1/dataset/1/",
        "directory": "",
        "filename": "testfile1.txt",
        "md5sum": "bogus",
        "mimetype": "text/plain",
        "modification_time": None,
        "parameter_sets": [
        ],
        "replicas": [
            {
                "datafile": "/api/v1/dataset_file/1/",
                "id": 1,
                "location": "local box at /home/mytardis/var/local",
                "resource_uri": "/api/v1/replica/1/",
                "uri": "subdir/testfile1.txt",
                "verified": True
            }
        ],
        "resource_uri": "/api/v1/dataset_file/1/",
        "size": 32
    }
    mock_datafile_get_response = json.dumps(mock_datafile)
    with requests_mock.Mocker() as mocker:
        get_datafile_url = "%s/api/v1/dataset_file/1/?format=json" % config.url
        mocker.get(get_datafile_url, text=mock_datafile_get_response)
        df_controller = DataFileController()
        args = Namespace(
            model='datafile', command='get', datafile_id=1, json=True,
            metadata=False, verbose=False)
        df_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_datafile


def test_datafile_get_cli_table(capfd):
    """
    Test getting datafile record via the command-line interface,
    requesting output in ASCII table format
    """
    mock_datafile = {
        "id": 1,
        "created_time": "2016-11-10T13:50:25.258483",
        "dataset": "/api/v1/dataset/1/",
        "directory": "",
        "filename": "testfile1.txt",
        "md5sum": "bogus",
        "mimetype": "text/plain",
        "modification_time": None,
        "parameter_sets": [
            {
                "id": 1,
                "datafile": "/api/v1/dataset_file/1/",
                "schema": {
                    "id": 1,
                    "name": "Series Metadata",
                    "namespace": "http://schema/namespace",
                    "hidden": False,
                    "immutable": True,
                    "subtype": "",
                    "type": 1
                },
                "parameters": [
                    {
                        "id": 1,
                        "name": "/api/v1/parametername/1/",
                        "parameterset": "/api/v1/datafileparameterset/1/",
                        "resource_uri": "/api/v1/datafileparameter/1/",
                        "string_value": "param value",
                        "numerical_value": None,
                        "datetime_value": None,
                        "link_id": None,
                        "value": None
                    }
                ]
            }
        ],
        "replicas": [
            {
                "datafile": "/api/v1/dataset_file/1/",
                "id": 1,
                "location": "local box at /home/mytardis/var/local",
                "resource_uri": "/api/v1/replica/1/",
                "uri": "subdir/testfile1.txt",
                "verified": True
            }
        ],
        "resource_uri": "/api/v1/dataset_file/1/",
        "size": 32
    }
    mock_datafile_get_response = json.dumps(mock_datafile)
    mock_schema = {
        "hidden": False,
        "id": 1,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": "",
        "type": 1
    }
    mock_schema_response = json.dumps(mock_schema)
    mock_pname = {
        "id": 1,
        "choices": "",
        "comparison_type": 1,
        "data_type": 2,
        "full_name": "Parameter Name",
        "immutable": True,
        "is_searchable": False,
        "name": "param name",
        "order": 1,
        "resource_uri": "/api/v1/parametername/1/",
        "schema": "/api/v1/schema/1/",
        "units": ""
    }
    mock_pname_response = json.dumps(mock_pname)
    expected = textwrap.dedent("""
        +----------------+---------------------------------------+
        | DataFile field |                 Value                 |
        +================+=======================================+
        | ID             | 1                                     |
        +----------------+---------------------------------------+
        | Dataset        | /api/v1/dataset/1/                    |
        +----------------+---------------------------------------+
        | Storage Box    | local box at /home/mytardis/var/local |
        +----------------+---------------------------------------+
        | Directory      |                                       |
        +----------------+---------------------------------------+
        | Filename       | testfile1.txt                         |
        +----------------+---------------------------------------+
        | URI            | subdir/testfile1.txt                  |
        +----------------+---------------------------------------+
        | Verified       | True                                  |
        +----------------+---------------------------------------+
        | Size           |  32 bytes                             |
        +----------------+---------------------------------------+
        | MD5 Sum        | bogus                                 |
        +----------------+---------------------------------------+

        +----------------------+-------------+----------------+--------------+-----------------+----------------+---------+
        | DataFileParameter ID |   Schema    | Parameter Name | String Value | Numerical Value | Datetime Value | Link ID |
        +======================+=============+================+==============+=================+================+=========+
        |                    1 | Schema Name | param name     | param value  |                 |                |         |
        +----------------------+-------------+----------------+--------------+-----------------+----------------+---------+
    """)
    with requests_mock.Mocker() as mocker:
        get_datafile_url = "%s/api/v1/dataset_file/1/?format=json" % config.url
        mocker.get(get_datafile_url, text=mock_datafile_get_response)
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_schema_response)
        get_pname_url = "%s/api/v1/parametername/1/?format=json" % config.url
        mocker.get(get_pname_url, text=mock_pname_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'datafile', 'get', '1', '--metadata']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv



"""
Controller base class for running commands (list, get, create, update)
"""
from __future__ import print_function

from ..utils import get_render_format
from ..views import render


class ModelCliController(object):
    """
    Controller base class for running commands (list, get, create, update)
    on model records

    Cli is an abbreviation for Command Line Interface
    """
    def __init__(self):
        self.allowed_commands = ["list", "get"]
        self.primary_key_arg = "id"
        self.model = None

    def run_command(self, args):
        """
        Generic run command method.
        """
        if args.command not in self.allowed_commands:
            raise NotImplementedError("Invalid command: %s" % args.command)
        render_format = get_render_format(args)
        getattr(self, args.command)(args, render_format)

    def list(self, args, render_format):
        """
        Display a list of records
        """
        if not self.model:
            return
        result_set = self.model.list(
            args.limit, args.offset, args.order_by)
        print(render(result_set, render_format))

    def get(self, args, render_format):
        """
        Display a single record
        """
        if not self.model:
            return
        primary_key = getattr(args, self.primary_key_arg)
        instance = self.model.objects.get(id=primary_key)
        print(render(instance, render_format))



"""
tests/argparse/test_invalid_model_cli.py
"""
import sys
import textwrap

import pytest

from mtclient.argparser import ArgParser


def test_invalid_model_argparse(capfd):
    """
    Test command-line interface with invalid model name
    """
    sys_argv = sys.argv
    sys.argv = ['mytardis', 'invalid_model', 'list']

    with pytest.raises(SystemExit) as err:
        ArgParser().get_args()
    assert err.value.code == 2

    expected = textwrap.dedent("""
         usage: mytardis [-h] [--verbose] [--version]
                         {api,config,version,facility,instrument,experiment,dataset,datafile,storagebox,schema}
                         ...
         mytardis: error: argument model: invalid choice: 'invalid_model' (choose from 'api', 'config', 'version', 'facility', 'instrument', 'experiment', 'dataset', 'datafile', 'storagebox', 'schema')
    """)
    _, err = capfd.readouterr()
    assert err.strip() == expected.strip()

    sys.argv = sys_argv



"""
This module contains the :class:`QuerySet` class, an abstraction to represent
a query to send to the MyTardis API.  It is designed to be lazy, so it will
only submit the query / queries to the REST API when we iterate through the
results.
"""


class QuerySet(object):
    """
    An abstraction to represent a query to send to the MyTardis API.  It is
    designed to be lazy, so it will only submit the query / queries to the
    REST API when we iterate through the results.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, model, filters=None, limit=None, offset=None,
                 order_by=None):
        """
        Each record in the query set can be
        represented as an object of class model
        """
        self.model = model
        self._filters = filters
        self._limit = limit
        self._offset = offset or 0
        self._order_by = order_by

        self._result_set = None

    def order_by(self, order_by):
        """
        Populate the self._order_by instance attribute
        """
        self._order_by = order_by
        return self

    def _execute_query(self):
        """
        The user has requested something which requires evaluating the query
        """
        self._result_set = self.model.list(
            filters=self._filters, limit=self._limit, offset=self._offset,
            order_by=self._order_by)

    def __repr__(self):
        """
        String representation
        """
        if not self._result_set:
            self._execute_query()
        pre_ellipsis = ""
        if self._result_set.offset > 0:
            pre_ellipsis = "..."
        post_ellipsis = ""
        if self._result_set.total_count > \
                len(self._result_set) + self._result_set.offset:
            post_ellipsis = "..."
        return "<QuerySet [%s%s%s]>" % (
            pre_ellipsis,
            ", ".join(str(obj) for obj in self._result_set),
            post_ellipsis)

    def __iter__(self):
        """
        Return an iterator for the QuerySet
        """
        if not self._result_set:
            self._execute_query()
        for index in range(0, len(self._result_set)):
            response_dict = \
                self._result_set.response_dict['objects'][index]
            yield self.model(response_dict)

        while self._result_set.next:
            self._offset += self._result_set.limit
            self._execute_query()
            for index in range(0, len(self._result_set)):
                response_dict = \
                    self._result_set.response_dict['objects'][index]
                yield self.model(response_dict)



"""
Model class for MyTardis API v1's StorageBoxResource.
"""
from __future__ import print_function

import logging
import requests

from ..conf import config
from .model import Model

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class StorageBox(Model):
    """
    Model class for MyTardis API v1's StorageBoxResource.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict):
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.name = response_dict['name']
        self.description = response_dict['description']
        self.django_storage_class = response_dict['django_storage_class']
        self.max_size = response_dict['max_size']
        self.status = response_dict['status']
        self.response_dict = response_dict
        self.attributes = []
        for attribute_json in response_dict['attributes']:
            self.attributes.append(StorageBoxAttribute(attribute_json))
        self.options = []
        for option_json in response_dict['options']:
            self.options.append(StorageBoxOption(option_json))

    def __str__(self):
        """
        Return a string representation of a storage box
        """
        return "<%s: %s>" % (type(self).__name__, self.name)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of storage boxes.

        :param filters: Filters, e.g. "id=123" or "name=box1"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`StorageBox` records, encapsulated in a
            `ResultSet` object`.
        """
        from ..utils import add_filters, extend_url
        from .resultset import ResultSet

        url = "%s/api/v1/storagebox/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(StorageBox, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single storage box record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the StorageBox to retrieve

        :return: A :class:`StorageBox` record.

        :raises requests.exceptions.HTTPError:
        """
        storage_box_id = kwargs.get("id")
        if not storage_box_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for StorageBox get "
                "at this stage.")
        url = "%s/api/v1/storagebox/%s/?format=json" % \
            (config.url, storage_box_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return StorageBox(response.json())


class StorageBoxAttribute(object):
    """
    Model class for MyTardis API v1's StorageBoxAttributeResource.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        self.key = response_dict['key']
        self.value = response_dict['value']
        self.response_dict = response_dict


class StorageBoxOption(object):
    """
    Model class for MyTardis API v1's StorageBoxOptionResource.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        self.key = response_dict['key']
        self.value = response_dict['value']
        self.response_dict = response_dict



"""
test_render_instrument.py

Tests for rendering a view of an instrument record
"""
import json
import textwrap

import requests_mock

from mtclient.conf import config
from mtclient.models.instrument import Instrument
from mtclient.views import render


def test_render_instrument():
    """
    Test rendering a view of an instrument record
    """
    mock_facility = {
        "id": 1,
        "manager_group": {
            "id": 1,
            "name": "test-facility-managers",
            "resource_uri": "/api/v1/group/1/"
        },
        "name": "Test Facility",
        "resource_uri": "/api/v1/facility/1/"
    }
    mock_facility_get_response = json.dumps(mock_facility)
    mock_instrument = {
        "created_time": None,
        "facility": {
            "created_time": None,
            "id": 1,
            "manager_group": {
                "id": 1,
                "name": "test-facility-managers",
                "resource_uri": "/api/v1/group/1/"
            },
            "modified_time": None,
            "name": "Test Facility",
            "resource_uri": "/api/v1/facility/1/"
        },
        "id": 1,
        "modified_time": None,
        "name": "Test Instrument",
        "resource_uri": "/api/v1/instrument/1/"
    }
    mock_instrument_get_response = json.dumps(mock_instrument)
    expected = textwrap.dedent("""
        +------------------+---------------------------+
        | Instrument field |           Value           |
        +==================+===========================+
        | ID               | 1                         |
        +------------------+---------------------------+
        | Name             | Test Instrument           |
        +------------------+---------------------------+
        | Facility         | <Facility: Test Facility> |
        +------------------+---------------------------+
    """)
    with requests_mock.Mocker() as mocker:
        get_facility_url = "%s/api/v1/facility/1/?format=json" % config.url
        mocker.get(get_facility_url, text=mock_facility_get_response)
        get_instrument_url = "%s/api/v1/instrument/1/?format=json" % config.url
        mocker.get(get_instrument_url, text=mock_instrument_get_response)

        instrument = Instrument.objects.get(id=1)

        assert render(instrument).strip() == expected.strip()



"""
Controller class for running commands (list)
on storage boxes.
"""
from ..models.storagebox import StorageBox
from .cli import ModelCliController


class StorageBoxController(ModelCliController):
    """
    Controller class for running commands (list, get) on storage box records
    """
    def __init__(self):
        super(StorageBoxController, self).__init__()
        self.allowed_commands = ["list", "get"]
        self.primary_key_arg = "storage_box_id"
        self.model = StorageBox



"""
tests/argparse/test_api_cli.py
"""
import sys

from mtclient.argparser import ArgParser


def test_api_argparse():
    """
    Test command-line interface for querying API endpoints
    """
    sys_argv = sys.argv
    sys.argv = ['mytardis', 'api', 'list']
    args = ArgParser().get_args()
    assert args.model == 'api'
    assert args.command == 'list'
    assert not args.json
    assert not args.verbose

    sys.argv = ['mytardis', 'api', 'get', 'dataset', '--json']
    args = ArgParser().get_args()
    assert args.model == 'api'
    assert args.command == 'get'
    assert args.json
    assert not args.verbose

    sys.argv = sys_argv



"""
tests/argparse/test_get_parser.py
"""
from mtclient.argparser import get_parser


def test_get_parser():
    """
    Test get_parser() method used by sphinx-argparse
    """
    parser = get_parser()
    assert parser.description == "Command-line interface for MyTardis REST API."
